# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


from enum import Enum

from .base import FeatureMap, FeatureMapConfig
from .softmax import NormDistribution, SMHyperbolic, SMOrf, SMReg


class FeatureMapType(str, Enum):
    SMOrf = "sm_orf"
    SMHyp = "sm_hyp"
    SMReg = "sm_reg"  # regularized softmax kernel


__all__ = [
    "SMOrf",
    "SMReg",
    "SMHyperbolic",
    "NormDistribution",
    "FeatureMapConfig",
    "FeatureMap",
]



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


from dataclasses import dataclass
from typing import Optional

import torch
import torch.nn as nn

from xformers.components.attention import (
    Attention,
    AttentionConfig,
    AttentionMask,
    maybe_sparsify,
    register_attention,
    sparsify,
)
from xformers.components.attention.attention_patterns import (
    causal_1d_pattern,
    random_pattern,
)
from xformers.components.attention.core import scaled_dot_product_attention


@dataclass
class RandomAttentionConfig(AttentionConfig):
    r: Optional[
        float
    ]  # the ratio of keys that the query can attend to. 1.0 means dense attention
    constant_masking: Optional[
        bool
    ]  # whether the randomness is per query or defined at construction time
    force_sparsity: Optional[bool]  # use sparsity in any case (potentially slower)


@register_attention("random", RandomAttentionConfig)
class RandomAttention(Attention):
    def __init__(
        self,
        dropout: float,
        causal: bool = False,
        r: float = 0.01,
        constant_masking: bool = True,
        force_sparsity: bool = False,
        *args,
        **kwargs,
    ):
        """
        "Random" attention, as proposed for instance in _BigBird.
        Random means in that case means that each query can attend to a random set of keys.
        This implementation is sparse-aware, meaning that the empty attention parts will not be represented in memory.

        Args:
            r (float): the ratio in [0,1] of keys that the query can attend to
            constant_masking (bool): if true, keep the same random set for all queries.

        _BigBird: "Zaheer, M., Guruganesh, G., Dubey, A., Ainslie, J., Alberti, C., Ontanon, S., Pham, P., Ravula,
         A., Wang, Q., Yang, L., & Ahmed, A. (2020). Big Bird: Transformers for Longer Sequences. ArXiv, NeurIPS."
        """
        super().__init__()

        self.attn_drop = nn.Dropout(dropout, inplace=False)
        self.causal = causal
        self.r = r
        self.rand_attention_mask: Optional[torch.Tensor] = None
        self.constant_masking = constant_masking
        self.force_sparsity = force_sparsity
        self.requires_same_k_q_dimensions = True

    def _get_rand_mask(self, shape: torch.Size) -> torch.Tensor:
        sparsity = 1 - self.r
        mask = random_pattern(shape[1], sparsity=sparsity)

        if self.causal:
            mask &= causal_1d_pattern(shape[1])

        mask = sparsify(mask) if self.force_sparsity else maybe_sparsify(mask)

        return mask

    def forward(
        self,
        q: torch.Tensor,
        k: torch.Tensor,
        v: torch.Tensor,
        att_mask: Optional[torch.Tensor] = None,
        *args,
        **kwargs,
    ):
        # Rand masking
        if not self.constant_masking or self.rand_attention_mask is None:
            self.rand_attention_mask = self._get_rand_mask(q.shape).to(q.device)

        # Mask-aware attention
        if att_mask is not None:
            if att_mask.dtype == torch.bool and isinstance(
                self.rand_attention_mask, AttentionMask
            ):
                mask = self.rand_attention_mask + AttentionMask.from_bool(att_mask)
            else:
                mask = self.rand_attention_mask & att_mask
        else:
            mask = self.rand_attention_mask

        # Handle q/k/v which would not fit the mask
        seq_len = q.shape[-2]
        q_, k_, v_ = map(lambda x: self._maybe_pad_sequence(x, mask), (q, k, v))

        # Normal attention with the random mask
        att = scaled_dot_product_attention(
            q=q_, k=k_, v=v_, att_mask=mask, dropout=self.attn_drop
        )

        # Take into account an hypothetical padding
        return att[:, :seq_len, :]



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.

import logging
from collections import namedtuple
from typing import Any, Dict, List

import matplotlib.pyplot as plt
import seaborn as sns
import torch

sns.set()

TestCase = namedtuple("TestCase", ["function", "name"])


_triton_is_available = torch.cuda.is_available()
if _triton_is_available:
    try:
        import triton
    except ImportError as e:
        logging.warning(f"Triton is not available: {e}.\nbench_functions")
        _triton_is_available = False


def pretty_print(results, title, units):
    """ Printout the contents of a dict as a human-readable and Markdown compatible array"""
    print(title)
    header = " Units: {:<45}".format(units)
    print("| " + header + "|" + "".join("{0:<20}|".format(k) for k in results.keys()))

    offset = len(header)
    print(
        "|-{}|".format("-" * offset)
        + "".join("{}|".format("-" * 20) for _ in results.keys())
    )

    workloads: Dict[str, Any] = {k: [] for v in results.values() for k in v.keys()}
    for v in results.values():
        for k in v.keys():
            workloads[k].append(v[k])

    for k, w in workloads.items():
        print(
            "| {0:<{offset}}|".format(k, offset=offset)
            + "".join("{:<20}|".format(v) for v in w)
        )

    print("")


def pretty_plot(results, title, units: str, filename=None, dash_key=""):
    """Graph out the contents of a dict.
    Dash key means that if the result label has this key, then it will be displayed with a dash"""

    if not filename:
        filename = title + ".png"

    # Sanitize the filename
    filename = (
        filename.replace(" ", "_").replace("/", "_").replace("-", "_").replace(":", "")
    )

    # Gather all the results in "collumns"
    workloads: Dict[str, Any] = {k: [] for v in results.values() for k in v.keys()}
    for v in results.values():
        for k in v.keys():
            workloads[k].append(float(v[k]))

    # Make sure that the plot is big enough
    f = plt.figure()
    f.set_figwidth(6)
    f.set_figheight(6)

    # Display the collections
    for k, v in workloads.items():
        if dash_key and dash_key in k:
            plt.plot(list(results.keys()), v, "--")
        else:
            plt.plot(list(results.keys()), v)

    plt.title(title)
    plt.legend(list(workloads.keys()), loc="lower right")
    plt.ylabel(units)
    plt.xticks(rotation=45)

    plt.savefig(filename, bbox_inches="tight")
    plt.close(f)


if _triton_is_available:

    def bench_functions(
        test_cases: List[TestCase], shapes, metric_transform, unit, title=""
    ):
        device = torch.device("cuda")

        for dtype in [torch.float16, torch.float32]:
            results: Dict[str, Any] = {}

            for B, M, K in shapes:
                a = torch.rand(B, M, K, device=device, dtype=dtype, requires_grad=True)

                for testcase in test_cases:
                    time = triton.testing.do_bench(lambda: testcase.function(a))[0]

                    metric = metric_transform(a, time)

                    key = f"B={B}, M={M}, K={K}"
                    if key not in results:
                        results[key] = {}

                    results[key][testcase.name] = f"{metric:.1f}"

            pretty_print(
                results,
                title=" ------------- Type: {} ------------- ".format(dtype),
                units=unit,
            )
            _type = " fp16" if dtype == torch.float16 else " fp32"
            pretty_plot(results, title + _type, unit, dash_key="pytorch")



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


from dataclasses import asdict, dataclass
from typing import Optional

import torch
import torch.nn as nn
from torch.nn.init import constant_

from xformers.components.attention import Attention
from xformers.components.in_proj_container import InProjContainer, InProjParams
from xformers.components.positional_embedding import RotaryEmbedding


@dataclass
class MultiHeadDispatchConfig:
    dim_model: int
    residual_dropout: float
    num_heads: int
    attention: Attention
    bias: bool
    dim_key: Optional[int]
    dim_value: Optional[int]
    in_proj_container: Optional[InProjContainer]
    use_separate_proj_weight: Optional[bool]
    use_rotary_embeddings: Optional[bool]
    out_proj: Optional[nn.Module]


# Move head forward and fold into batch dim. dimensions become (B * nh, S, hs)
def _fold_heads(t: torch.Tensor, B: int, S: int, H: int, Hs: int):
    return t.view(B, S, H, Hs).transpose(1, 2).flatten(start_dim=0, end_dim=1)


def _split_heads(t: torch.Tensor, B: int, S: int, H: int, Hs: int):
    return t.view(B, S, H, Hs).transpose(1, 2)


class MultiHeadDispatch(nn.Module):
    """
    A multi-head masked self-attention dispatch mechanism, with a projection at the end,
    following the architecture proposed in `Attention is all you need`_, Vaswani et al.

    The actual attention mechanism can vary, as well as the projections.
    This can be used to wrap the proposed attention mechanisms and make them multi-head aware,
    but it is optional.

    .. _`Attention is all you need`: https://arxiv.org/abs/1706.03762v5
    """

    def __init__(
        self,
        dim_model: int,
        residual_dropout: float,
        num_heads: int,
        attention: Attention,
        bias: bool = True,
        dim_key: Optional[int] = None,
        dim_value: Optional[int] = None,
        in_proj_container: Optional[InProjContainer] = None,
        use_separate_proj_weight: Optional[bool] = False,
        use_rotary_embeddings: Optional[bool] = False,
        out_proj: Optional[nn.Module] = None,
        *args,
        **kwargs,
    ):
        super().__init__()

        assert (
            dim_model % num_heads == 0
        )  # static preset for now, each head works on 1/d the embeddings, could be relaxed
        assert num_heads > 0

        # Popular default is that all latent dimensions are the same
        dim_key, dim_value = map(lambda x: x if x else dim_model, (dim_key, dim_value))

        self.num_heads = num_heads
        self.dim_k = dim_key // num_heads
        self.dim_value = dim_value
        self.dim_model = dim_model
        self.attention = attention

        # key, query, value projections for all heads
        # critical options are
        # - are we sharing weights ?
        # - are we adding biases, and if yes are they shared ?
        if attention.requires_input_projection:
            self.in_proj_container = (
                in_proj_container
                if in_proj_container is not None
                else InProjContainer(
                    query_proj_params=InProjParams(dim_model, dim_key, bias=bias),
                    key_proj_params=InProjParams(dim_model, dim_key, bias=bias)
                    if use_separate_proj_weight
                    else None,
                    value_proj_params=InProjParams(dim_model, dim_value, bias=bias)
                    if use_separate_proj_weight
                    else None,
                )
            )

        # Optional rotary embeddings
        self.rotary_embeddings = (
            RotaryEmbedding(self.dim_k) if use_rotary_embeddings else None
        )

        # Regularization
        self.resid_drop = nn.Dropout(residual_dropout, inplace=False)

        # Output projection
        self.proj = out_proj if out_proj else nn.Linear(dim_model, dim_model, bias=bias)
        if isinstance(self.proj, nn.Linear) and self.proj.bias is not None:
            constant_(self.proj.bias, 0.0)

    def _check(self, t, name):
        assert (
            t.shape[2] % self.dim_k == 0
        ), f"the {name} embeddings need to be divisible by the number of heads"

    def forward(
        self,
        query: torch.Tensor,
        key: Optional[torch.Tensor] = None,
        value: Optional[torch.Tensor] = None,
        att_mask: Optional[torch.Tensor] = None,
        key_padding_mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """
        Expected input dimensions are [batch size, sequence length, embed dim]
        Output dimensions are [batch size, sequence length, embed dim]
        """

        if key is None:
            key = query
        if value is None:
            value = query

        # Check the dimensions properly
        self._check(query, "query")
        self._check(value, "value")
        self._check(key, "key")

        max_batch = max((query.shape[0], key.shape[0], value.shape[0]))
        query, key, value = map(
            lambda x: x.expand(max_batch, -1, -1), [query, key, value]
        )

        B, S_Q, _ = query.size()  # Batch x Sequence x Embedding (latent)
        _, S_K, _ = key.size()  # K, Q's sequence length could differ

        # Catch different query and key length but a causal attention
        if S_Q != S_K:
            assert (
                not self.attention.requires_same_k_q_dimensions
            ), "This attention mechanism requires query and key to have the same sequence (context) lengths"

            if hasattr(self.attention, "causal"):
                assert not self.attention.causal, (
                    "Causal attention is not supported when key and query have different sequence lengths.\n"
                    + "In that case causality is ill-determined. Please pad your sequences accordingly"
                )

        if self.attention.requires_skip_multi_head:
            return self.attention(
                query, key, value, att_mask=att_mask, key_padding_mask=key_padding_mask
            )

        # Calculate query, key, values for all heads in batch
        if self.attention.requires_input_projection:
            q, k, v = self.in_proj_container(query=query, key=key, value=value)
        else:
            k, q, v = key, query, value

        # Optional: rotary embedding, add relative positioning information
        if self.rotary_embeddings:
            # rotary requires the head dimension
            q = _split_heads(q, B, S_Q, self.num_heads, self.dim_k)
            k = _split_heads(k, B, S_K, self.num_heads, self.dim_k)
            v = _split_heads(v, B, S_K, self.num_heads, self.dim_k)

            q, k = self.rotary_embeddings(q=q, k=k)

            if not self.attention.requires_head_dimension:
                q, k, v = q.flatten(0, 1), k.flatten(0, 1), v.flatten(0, 1)

        else:
            # Reshape k/q/v to either expose the heads, or fold the head dimension into the batch
            reshape_fn = (
                _split_heads if self.attention.requires_head_dimension else _fold_heads
            )

            q = reshape_fn(q, B, S_Q, self.num_heads, self.dim_k)
            k = reshape_fn(k, B, S_K, self.num_heads, self.dim_k)
            v = reshape_fn(v, B, S_K, self.num_heads, self.dim_k)

        # Self-attend
        y = self.attention(
            q=q, k=k, v=v, att_mask=att_mask, key_padding_mask=key_padding_mask
        )

        # Re-assemble all head outputs side by side
        y = (
            y.view(B, self.num_heads, S_Q, self.dim_k)
            .transpose(1, 2)
            .flatten(start_dim=2, end_dim=3)
        )

        # Output projection, dropout and good to go
        y = self.resid_drop(self.proj(y))

        # Return the same sequence size as the input
        return y

    @classmethod
    def from_config(cls, config: MultiHeadDispatchConfig):
        # Generate the class inputs from the config
        fields = asdict(config)

        # Skip all Nones so that default values are used
        fields = {k: v for k, v in fields.items() if v is not None}

        return cls(**fields)



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


import torch
import triton
import triton.language as tl
from triton.ops.matmul_perf_model import estimate_matmul_time, prune_num_stages

# This implements a simple QKt matrix multiplication (non ragged), for reference
# Author: Tom B Brown


def init_to_zero(name):
    return lambda nargs: nargs[name].zero_()


def get_configs_io_bound():
    configs = []
    for num_stages in [2, 3, 4, 5, 6]:
        for block_m in [16, 32]:
            for block_k in [32, 64]:
                for block_n in [32, 64, 128, 256]:
                    num_warps = 2 if block_n <= 64 else 4
                    configs.append(
                        triton.Config(
                            {
                                "BLOCK_M": block_m,
                                "BLOCK_K": block_n,
                                "BLOCK_D": block_k,
                            },
                            num_stages=num_stages,
                            num_warps=num_warps,
                        )
                    )

    return configs


def get_all_configs():
    return [
        # basic configs for compute-bound matmuls
        triton.Config(
            {"BLOCK_M": 128, "BLOCK_K": 256, "BLOCK_D": 32},
            num_stages=3,
            num_warps=8,
        ),
        triton.Config(
            {"BLOCK_M": 256, "BLOCK_K": 128, "BLOCK_D": 32},
            num_stages=3,
            num_warps=8,
        ),
        triton.Config(
            {"BLOCK_M": 256, "BLOCK_K": 64, "BLOCK_D": 32},
            num_stages=4,
            num_warps=4,
        ),
        triton.Config(
            {"BLOCK_M": 64, "BLOCK_K": 256, "BLOCK_D": 32},
            num_stages=4,
            num_warps=4,
        ),
        triton.Config(
            {"BLOCK_M": 128, "BLOCK_K": 128, "BLOCK_D": 32},
            num_stages=4,
            num_warps=4,
        ),
        triton.Config(
            {"BLOCK_M": 128, "BLOCK_K": 64, "BLOCK_D": 32},
            num_stages=4,
            num_warps=4,
        ),
        triton.Config(
            {"BLOCK_M": 64, "BLOCK_K": 128, "BLOCK_D": 32},
            num_stages=4,
            num_warps=4,
        ),
        triton.Config(
            {"BLOCK_M": 128, "BLOCK_K": 32, "BLOCK_D": 32},
            num_stages=4,
            num_warps=4,
        ),
        triton.Config(
            {"BLOCK_M": 64, "BLOCK_K": 32, "BLOCK_D": 32},
            num_stages=5,
            num_warps=2,
        ),
    ] + get_configs_io_bound()


def get_fast_dev_configs():
    return [
        triton.Config(
            {"BLOCK_Q": 64, "BLOCK_K": 32, "BLOCK_D": 32},
            num_stages=5,
            num_warps=2,
        )
    ]


# fmt: off
@triton.autotune(
    # configs=get_all_configs(),
    configs=get_fast_dev_configs(),
    key=["n_ctx_q", "n_ctx_k", "d_model"],
    prune_configs_by={
        "prune_num_stages_by": prune_num_stages,
        "perf_model": estimate_matmul_time,
        "top_k": 10,
    },
)
@triton.jit
def _kernel(
    q_ptr, k_ptr, scores_ptr,
    n_ctx_q,
    n_ctx_k,  # N
    d_model,
    stride_ctx_q, stride_ctx_k,
    stride_d,  # Stride along the d_model_per_head dim
    stride_out_q, stride_out_k,
    BLOCK_Q: tl.constexpr,
    BLOCK_K: tl.constexpr,
    BLOCK_D: tl.constexpr,
):
    # fmt: on

    # matrix multiplication
    pid = tl.program_id(0)

    # Determine the number of blocks in the grid
    grid_k = (n_ctx_k + BLOCK_K - 1) // BLOCK_K

    pid_q = pid // grid_k
    pid_k = pid % grid_k

    # do matrix multiplication
    rq = pid_q * BLOCK_Q + tl.arange(0, BLOCK_Q)
    rq = tl.max_contiguous(tl.multiple_of(rq % n_ctx_q, BLOCK_Q), BLOCK_Q)

    rk = pid_k * BLOCK_K + tl.arange(0, BLOCK_K)
    rk = tl.max_contiguous(tl.multiple_of(rk % n_ctx_k, BLOCK_K), BLOCK_K)

    # Iterate through blocks of the d_model dimension and accumulate values into acc
    acc_tile = tl.zeros((BLOCK_Q, BLOCK_K), dtype=tl.float32)
    rd = tl.arange(0, BLOCK_D)

    q_ptr_tile = q_ptr + (rq[:, None] * stride_ctx_q + rd[None, :] * stride_d)
    k_ptr_tile = k_ptr + (rd[:, None] * stride_d + rk[None, :] * stride_ctx_k)

    for d_max_offset in range(d_model, 0, -BLOCK_D):
        q_tile = tl.load(q_ptr_tile, mask=rd[None, :] < d_max_offset, other=0.0)
        k_tile = tl.load(k_ptr_tile, mask=rd[:, None] < d_max_offset, other=0.0)

        # In einsum notation, the following does: qd,dk->qk
        acc_tile += tl.dot(q_tile, k_tile)

        q_ptr_tile += BLOCK_D * stride_d
        k_ptr_tile += BLOCK_D * stride_d

    acc_tile = acc_tile.to(scores_ptr.dtype.element_ty)

    # We rematerialize rq and rk here because it allows them to be deallocated above
    # instead of being kept in registers throughout the inner for-loop
    rq = pid_q * BLOCK_Q + tl.arange(0, BLOCK_Q)
    rk = pid_k * BLOCK_K + tl.arange(0, BLOCK_K)

    scores_offset_tile = rq[:, None] * stride_out_q + rk[None, :] * stride_out_k
    scores_ptr_tile = scores_ptr + scores_offset_tile

    mask = (rq < n_ctx_q)[:, None] & (rk < n_ctx_k)[None, :]

    tl.store(scores_ptr_tile, acc_tile, mask=mask)


def qk_dotprod(query, key):
    device = query.device

    # handle non-contiguous inputs if necessary
    if query.stride(0) > 1 and query.stride(1) > 1:
        query = query.contiguous()
    if key.stride(0) > 1 and key.stride(1) > 1:
        key = key.contiguous()

    # check constraints
    n_ctx_q, d_model = query.shape
    n_ctx_k, d_model_k = key.shape
    assert d_model == d_model_k, f"{query.shape=} {key.shape=}"

    # allocates output
    scores_out = torch.empty((n_ctx_q, n_ctx_k), device=device, dtype=query.dtype)

    # Stride along the d_model dimension
    stride_d = query.stride(1)
    assert stride_d == key.stride(1), f"{stride_d=}, {key.stride(1)=}"

    # launch kernel
    def grid(META):
        return (
            triton.cdiv(n_ctx_q, META["BLOCK_Q"])
            * triton.cdiv(n_ctx_k, META["BLOCK_K"]),
        )

    _kernel[grid](
        query,
        key,
        scores_out,
        n_ctx_q,
        n_ctx_k,
        d_model,
        query.stride(0),  # stride_ctx_q
        key.stride(0),  # stride_ctx_k
        stride_d,  # stride_d
        scores_out.stride(0),  # stride_out_q
        scores_out.stride(1),  # stride_out_k
    )
    return scores_out



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


from functools import lru_cache
from typing import List, Tuple

import torch
from ragged_inference.garbage_pad_ragged_acts import RaggedActivations


class SingleSeqKVCache:
    def __init__(self, keys: torch.Tensor, values: torch.Tensor):
        # Tensor of shape [2, n_ctx, d_model_per_gpu]
        # - keys are cache[0]
        # - values are cache[1]
        self.raw_keys = keys
        self.raw_values = values

    @property
    def keys(self) -> torch.Tensor:
        return self.raw_keys

    @property
    def values(self) -> torch.Tensor:
        return self.raw_values

    @property
    def n_ctx(self):
        return self.raw_values.shape[0]

    @property
    def d_model_per_gpu(self):
        return self.raw_values.shape[-1]

    @property
    def is_cuda(self):
        return self.raw_values.is_cuda

    @property
    def dtype(self):
        return self.raw_values.dtype


def extend_kv_caches(
    seq_kv_cache: List[SingleSeqKVCache],
    active_keys: RaggedActivations,
    active_values: RaggedActivations,
) -> List[SingleSeqKVCache]:
    assert seq_kv_cache[0].is_cuda

    updated_seq_kv_cache = []
    for cache, keys, values in zip(
        seq_kv_cache, active_keys.iter_full_tensors(), active_values.iter_full_tensors()
    ):

        # Dim 1 is the context
        new_cache = SingleSeqKVCache(
            keys=torch.cat([cache.keys, keys], dim=0),
            values=torch.cat([cache.values, values], dim=0),
        )
        updated_seq_kv_cache.append(new_cache)

    return updated_seq_kv_cache


def garbage_pad_seq_kv_cache(
    seq_kv_cache: List[SingleSeqKVCache],
) -> Tuple[torch.Tensor, torch.Tensor]:
    assert seq_kv_cache[0].is_cuda
    dtype = seq_kv_cache[0].dtype
    n_ctx_per_kv_cache = [seq.n_ctx for seq in seq_kv_cache]

    # Create a view so that the output is (n_seqs, n_ctx_max, d_model)
    # This should not incur an extra memcopy
    n_seqs = len(n_ctx_per_kv_cache)
    n_ctx_max = max(n_ctx_per_kv_cache)

    padded_keys = torch.empty(
        n_seqs,
        n_ctx_max,
        seq_kv_cache[0].d_model_per_gpu,
        dtype=dtype,
        device="cuda",
    )

    padded_values = torch.empty(
        n_seqs,
        n_ctx_max,
        seq_kv_cache[0].d_model_per_gpu,
        dtype=dtype,
        device="cuda",
    )

    for seq_idx, seq in enumerate(seq_kv_cache):
        padded_keys[seq_idx, : seq.n_ctx, :] = seq.keys
        padded_values[seq_idx, : seq.n_ctx, :] = seq.values
    return (padded_keys, padded_values)


def garbage_pad_keys(
    seq_kv_cache: List[SingleSeqKVCache],
) -> torch.Tensor:
    assert seq_kv_cache[0].is_cuda
    dtype = seq_kv_cache[0].dtype
    n_ctx_per_kv_cache = [seq.n_ctx for seq in seq_kv_cache]

    # Create a view so that the output is (n_seqs, n_ctx_max, d_model)
    # This should not incur an extra memcopy
    n_seqs = len(n_ctx_per_kv_cache)
    n_ctx_max = max(n_ctx_per_kv_cache)

    padded_keys = torch.empty(
        n_seqs,
        n_ctx_max,
        seq_kv_cache[0].d_model_per_gpu,
        dtype=dtype,
        device="cuda",
    )

    for seq_idx, seq in enumerate(seq_kv_cache):
        padded_keys[seq_idx, : seq.n_ctx, :] = seq.keys
    return padded_keys


@lru_cache(maxsize=1)  # Memoize because we repeat this for consecutive resblocks
def _create_indices(n_ctx_per_kv_cache):
    """
    We cache this because it requires some substantial CPU work and it's done multiple
    times sequentially (once per resblock)
    """
    indices_list = []
    ragged_idx = 0
    max_n_ctx = max(n_ctx_per_kv_cache)
    for n_ctx in n_ctx_per_kv_cache:
        for idx_into_seq in range(max_n_ctx):
            if idx_into_seq < n_ctx:
                indices_list.append(ragged_idx)
                ragged_idx += 1
            else:
                indices_list.append(0)  # Add a placeholder
    return torch.tensor(indices_list, device="cuda")


def calculate_scores_via_qk_dotprod(
    seq_kv_cache: List[SingleSeqKVCache],  # These have already been extended
    active_queries: RaggedActivations,
) -> torch.Tensor:
    padded_keys = garbage_pad_keys(seq_kv_cache)
    padded_active_queries = active_queries.to_garbage_padded()
    return torch.einsum("bkd,bqd->bqk", padded_keys, padded_active_queries)


def scores_via_qk_dotprod(
    query: RaggedActivations,
    key: RaggedActivations,
) -> torch.Tensor:
    padded_query = query.to_garbage_padded()
    padded_key = key.to_garbage_padded()
    return torch.einsum("bkd,bqd->bqk", padded_key, padded_query)



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.

import itertools

import torch
from torch.utils import benchmark

from xformers.components.attention.core import (
    SparseCS,
    _create_random_sparsity,
    _matmul_with_mask,
    _softmax,
    bmm,
)

MIN_RUN_TIME = 1
SHAPES = [[8, 8], [256, 1024], [128, 256]]
SPARSITIES = [0.5, 0.8, 0.9, 0.95, 0.99]


def bench_sddmm():
    min_run_time = MIN_RUN_TIME
    SPARSITIES = [0.95, 0.98, 0.99, 0.995, 0.999]

    device = torch.device("cuda")
    results = []

    for B, M, K in zip(*SHAPES):
        a = torch.rand(B, M, K, device=device)
        b = torch.rand(B, M, K, device=device)

        for backend, prob in itertools.product(
            ["coo_pytorch", "csr_sputnik", "csr_ge"], SPARSITIES
        ):
            mask = _create_random_sparsity(torch.ones(B, M, M, dtype=torch.bool), prob)
            aa = a
            bb = b
            if "csr" in backend:
                mask = SparseCS(mask, device)
                aa = a
                bb = b
                row_indices = mask.row_indices
                row_offsets = mask.row_offsets
                column_indices = mask.column_indices
                if "_ge" in backend:
                    fn = torch.ops.xformers.csr_sddmm
                else:
                    fn = torch.ops.xformers.sddmm_sputnik
                fn_str = "fn(a, b, row_indices, row_offsets, column_indices)"
            else:
                mask = mask.to_sparse().to(device)
                _, row_offsets, column_indices = mask.indices().int().unbind()
                row_offsets = row_offsets.contiguous()
                column_indices = column_indices.contiguous()
                row_indices = row_offsets

                bb = b.transpose(-2, -1)
                fn = _matmul_with_mask
                fn_str = "fn(a, b, mask)"

            results.append(
                benchmark.Timer(
                    stmt=fn_str,
                    globals={
                        "a": aa,
                        "b": bb,
                        "mask": mask,
                        "row_indices": row_indices,
                        "row_offsets": row_offsets,
                        "column_indices": column_indices,
                        "fn": fn,
                    },
                    label="sddmm",
                    sub_label=f"sparsity {backend}: {prob:0.4f}",
                    description=f"B={B}, M={M}, K={K}",
                ).blocked_autorange(min_run_time=min_run_time)
            )

    compare = benchmark.Compare(results)
    compare.print()


def bench_matmul_with_mask():
    min_run_time = MIN_RUN_TIME
    prob = 0.9
    device = torch.device("cuda")
    results = []

    for B, M, K in zip(*SHAPES):
        a = torch.rand(B, M, K, device=device)
        b = torch.rand(B, K, M, device=device)
        mask = torch.rand(B, M, M, device=device) > prob

        results.extend(
            [
                benchmark.Timer(
                    stmt="_matmul_with_mask(a, b, mask)",
                    globals={
                        "a": a,
                        "b": b,
                        "mask": None,
                        "_matmul_with_mask": _matmul_with_mask,
                    },
                    label="matmul_with_mask",
                    sub_label="dense",
                    description=f"B={B}, M={M}, K={K}",
                ).blocked_autorange(min_run_time=min_run_time),
                benchmark.Timer(
                    stmt="_matmul_with_mask(a, b, mask)",
                    globals={
                        "a": a,
                        "b": b,
                        "mask": mask,
                        "_matmul_with_mask": _matmul_with_mask,
                    },
                    label="matmul_with_mask",
                    sub_label="dense with masking",
                    description=f"B={B}, M={M}, K={K}",
                ).blocked_autorange(min_run_time=min_run_time),
            ]
        )
        for sputnik, prob in itertools.product([False, True], SPARSITIES):
            mask = _create_random_sparsity(
                torch.ones(B, M, M, dtype=torch.bool, device=device), prob
            )
            aa = a
            bb = b
            if sputnik:
                mask = SparseCS(mask, device)
                aa = a
                bb = b.transpose(-2, -1).contiguous().transpose(-2, -1)
            else:
                mask = mask.to_sparse()
            results.append(
                benchmark.Timer(
                    stmt="_matmul_with_mask(a, b, mask)",
                    globals={
                        "a": aa,
                        "b": bb,
                        "mask": mask,
                        "_matmul_with_mask": _matmul_with_mask,
                    },
                    label="matmul_with_mask",
                    sub_label=f"sparsity {'sputnik' if sputnik else 'pytorch'}: {prob:0.2f}",
                    description=f"B={B}, M={M}, K={K}",
                ).blocked_autorange(min_run_time=min_run_time)
            )

    compare = benchmark.Compare(results)
    compare.print()


def bench_softmax():
    min_run_time = MIN_RUN_TIME
    prob = 0.9
    device = torch.device("cuda")
    results = []

    for B, M, K in zip(*SHAPES):
        a = torch.rand(B, M, M, device=device)
        a[a < prob] = 0

        results.extend(
            [
                benchmark.Timer(
                    stmt="_softmax(a)",
                    globals={
                        "a": a,
                        "_softmax": _softmax,
                    },
                    label="softmax",
                    sub_label="dense",
                    description=f"B={B}, M={M}, K={K}",
                ).blocked_autorange(min_run_time=min_run_time),
            ]
        )
        for sputnik, prob in itertools.product([False, True], SPARSITIES):
            a = _create_random_sparsity(torch.rand(B, M, M, device=device), prob)
            if sputnik:
                a = SparseCS(a, device)
            else:
                a = a.to_sparse()
            results.append(
                benchmark.Timer(
                    stmt="_softmax(a)",
                    globals={
                        "a": a,
                        "_softmax": _softmax,
                    },
                    label="softmax",
                    sub_label=f"sparsity {'sputnik' if sputnik else 'pytorch'}: {prob:0.2f}",
                    description=f"B={B}, M={M}, K={K}",
                ).blocked_autorange(min_run_time=min_run_time)
            )

    compare = benchmark.Compare(results)
    compare.print()


def bench_bmm():
    min_run_time = MIN_RUN_TIME
    prob = 0.9
    device = torch.device("cuda")
    results = []

    for B, M, K in zip(*SHAPES):
        a = torch.rand(B, M, M, device=device)
        a[a < prob] = 0
        b = torch.rand(B, M, K, device=device)

        results.extend(
            [
                benchmark.Timer(
                    stmt="bmm(a, b)",
                    globals={
                        "a": a,
                        "b": b,
                        "bmm": bmm,
                    },
                    label="bmm",
                    sub_label="dense",
                    description=f"B={B}, M={M}, K={K}",
                ).blocked_autorange(min_run_time=min_run_time),
            ]
        )
        for sputnik, prob in itertools.product([False, True], SPARSITIES):
            a = _create_random_sparsity(torch.rand(B, M, M, device=device), prob)
            bb = b
            if sputnik:
                a = SparseCS(a, device)
                bb = b
            else:
                a = a.to_sparse()
            results.append(
                benchmark.Timer(
                    stmt="bmm(a, b)",
                    globals={
                        "a": a,
                        "b": bb,
                        "bmm": bmm,
                    },
                    label="bmm",
                    sub_label=f"sparsity {'sputnik' if sputnik else 'pytorch'}: {prob:0.2f}",
                    description=f"B={B}, M={M}, K={K}",
                ).blocked_autorange(min_run_time=min_run_time)
            )

    compare = benchmark.Compare(results)
    compare.print()


bench_sddmm()
bench_matmul_with_mask()
bench_softmax()
bench_bmm()



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.

from dataclasses import dataclass
from typing import List, Optional

import torch
import triton
import triton.language as tl
from ragged_inference.garbage_pad_ragged_acts import RaggedActivations
from triton.ops.matmul_perf_model import estimate_matmul_time, prune_num_stages

# This implements a ragged attention (batched attention mechanism natively handling different sequence sizes)
# Author: Tom B Brown


def init_to_zero(name):
    return lambda nargs: nargs[name].zero_()


# TODO: tune these
BLOCK_Q = 16
BLOCK_K = 128
BLOCK_D = 32


def get_fast_dev_configs():
    return [
        triton.Config(
            {"BLOCK_Q": BLOCK_Q, "BLOCK_K": BLOCK_K, "BLOCK_D": BLOCK_D},
            num_stages=5,
            num_warps=2,
        )
    ]


@triton.autotune(
    # configs=get_all_configs(),
    configs=get_fast_dev_configs(),
    key=["max_n_ctx_q_across_seqs", "max_n_ctx_k_across_seqs", "d_head"],
    prune_configs_by={
        "prune_num_stages_by": prune_num_stages,
        "perf_model": estimate_matmul_time,
        "top_k": 10,
    },
)
@triton.jit
def _qk_dotprod_kernel(
    # Pointers to our tensors
    q_ptr,
    k_ptr,
    scores_ptr,  # Rectangular output tensor
    # Pointers to lookup tables (sometimes referred to as a "lut")
    pid_to_in_q_token_offset_ptr,
    pid_to_in_k_token_offset_ptr,
    pid_to_out_q_block_ptr,
    pid_to_out_k_block_ptr,
    pid_to_out_seq_idx_ptr,
    # Integers
    max_n_ctx_q_across_seqs,
    max_n_ctx_k_across_seqs,
    d_head,
    stride_ctx_q,
    stride_ctx_k,
    stride_out_q,
    stride_out_k,
    stride_out_seq,
    total_ctx_q_across_all_seqs,
    total_ctx_k_across_all_seqs,
    # These get populated from the triton.Config
    BLOCK_Q: tl.constexpr,
    BLOCK_K: tl.constexpr,
    BLOCK_D: tl.constexpr,
):
    """
    Adapted from https://github.com/openai/triton/blob/v2.0/python/triton/ops/matmul.py
    """

    # matrix multiplication
    pid = tl.program_id(0)

    out_q_block = tl.load(pid_to_out_q_block_ptr + pid)
    out_k_block = tl.load(pid_to_out_k_block_ptr + pid)
    out_seq_idx = tl.load(pid_to_out_seq_idx_ptr + pid)
    in_q_token_offset = tl.load(pid_to_in_q_token_offset_ptr + pid)
    in_k_token_offset = tl.load(pid_to_in_k_token_offset_ptr + pid)

    # Define indices ranges, we follow the triton convention of prefixing
    # with "r" to denote a range like "rq" is the range for queries below
    rq = in_q_token_offset + tl.arange(0, BLOCK_Q)
    rk = in_k_token_offset + tl.arange(0, BLOCK_K)

    # Prevent out of bounds reads. It's ok to read garbage data for queries and keys that aren't
    # actually used. Their values don't effect any of the outputs.
    q_ctx_in_bounds = rq < total_ctx_q_across_all_seqs
    k_ctx_in_bounds = rk < total_ctx_k_across_all_seqs

    # We will accumulate all the d_head items into acc
    acc_tile = tl.zeros((BLOCK_Q, BLOCK_K), dtype=tl.float32)

    rd = tl.arange(0, BLOCK_D)  # rd indexes into the d_head dimension

    # We use broadcasting to convert our 1D ranges into 2D tiles
    q_ptr_tile = q_ptr + (rq[:, None] * stride_ctx_q + rd[None, :])
    k_ptr_tile = k_ptr + (rd[:, None] + rk[None, :] * stride_ctx_k)

    # After we have profiling, see if we can rewrite this to be more readable by
    # just updating rd += BLOCK_D
    for d_max_offset in range(d_head, 0, -BLOCK_D):
        q_tile = tl.load(
            q_ptr_tile,
            mask=(rd[None, :] < d_max_offset) & q_ctx_in_bounds[:, None],
            other=0.0,
        )
        k_tile = tl.load(
            k_ptr_tile,
            mask=(rd[:, None] < d_max_offset) & k_ctx_in_bounds[None, :],
            other=0.0,
        )

        # In einsum notation, the tl.dot does: qd,dk->qk
        # This should use tensorcores, so the inputs might be fp16, but the outputs
        # and all the internal accumulators are fp32
        acc_tile += tl.dot(q_tile, k_tile)

        q_ptr_tile += BLOCK_D
        k_ptr_tile += BLOCK_D

    # Figure out the output blocks
    rq_out = out_q_block * BLOCK_Q + tl.arange(0, BLOCK_Q)
    rk_out = out_k_block * BLOCK_K + tl.arange(0, BLOCK_K)

    scores_offset_tile = (
        rq_out[:, None] * stride_out_q
        + rk_out[None, :] * stride_out_k
        + out_seq_idx * stride_out_seq
    )
    scores_ptr_tile = scores_ptr + scores_offset_tile

    mask = (rq_out < max_n_ctx_q_across_seqs)[:, None] & (
        rk_out < max_n_ctx_k_across_seqs
    )[None, :]

    # Cast back to lower precision immediately before storing
    acc_tile = acc_tile.to(scores_ptr.dtype.element_ty)
    tl.store(scores_ptr_tile, acc_tile, mask=mask)


@dataclass
class RaggedQkPidLookupTable:
    # TODO: link to a drawing of what these tensors are
    # All cuda tensors
    pid_to_in_q_token_offset: torch.Tensor
    pid_to_in_k_token_offset: torch.Tensor
    pid_to_out_q_block: torch.Tensor
    pid_to_out_k_block: torch.Tensor
    pid_to_out_seq_idx: torch.Tensor
    n_pids_total: int

    @staticmethod
    def from_single_seq(n_ctx_q: int, n_ctx_k: int) -> "RaggedQkPidLookupTable":
        grid_q = triton.cdiv(n_ctx_q, BLOCK_Q)
        grid_k = triton.cdiv(n_ctx_k, BLOCK_K)
        n_pids_total = grid_q * grid_k

        pid_to_in_q_token_offset = torch.zeros(
            n_pids_total, dtype=torch.int32, device="cuda"
        )
        pid_to_in_k_token_offset = torch.zeros(
            n_pids_total, dtype=torch.int32, device="cuda"
        )
        pid_to_out_q_block = torch.zeros(n_pids_total, dtype=torch.int32, device="cuda")
        pid_to_out_k_block = torch.zeros(n_pids_total, dtype=torch.int32, device="cuda")
        pid_to_out_seq_idx = torch.zeros(n_pids_total, dtype=torch.int32, device="cuda")

        for pid in range(n_pids_total):
            q_block_idx = pid // grid_k
            k_block_idx = pid % grid_k

            in_q_token_offset = q_block_idx * BLOCK_Q
            in_k_token_offset = k_block_idx * BLOCK_K

            pid_to_out_q_block[pid] = q_block_idx
            pid_to_out_k_block[pid] = k_block_idx
            pid_to_in_q_token_offset[pid] = in_q_token_offset
            pid_to_in_k_token_offset[pid] = in_k_token_offset

        return RaggedQkPidLookupTable(
            pid_to_in_q_token_offset=pid_to_in_q_token_offset,
            pid_to_in_k_token_offset=pid_to_in_k_token_offset,
            pid_to_out_q_block=pid_to_out_q_block,
            pid_to_out_k_block=pid_to_out_k_block,
            pid_to_out_seq_idx=pid_to_out_seq_idx,
            n_pids_total=n_pids_total,
        )

    @staticmethod
    def from_query_and_key_tokens_per_seq(
        n_ctx_q_per_seq: List[int],
        n_ctx_k_per_seq: List[int],
        block_q_override: Optional[int] = None,
        block_k_override: Optional[int] = None,
    ) -> "RaggedQkPidLookupTable":
        block_q = block_q_override if block_q_override else BLOCK_Q
        block_k = block_k_override if block_k_override else BLOCK_K

        pid_to_in_q_token_offset = []
        pid_to_in_k_token_offset = []
        pid_to_out_q_block = []
        pid_to_out_k_block = []
        pid_to_out_seq_idx = []

        n_in_q_token_so_far = 0
        n_in_k_token_so_far = 0

        for seq_idx, (n_ctx_q, n_ctx_k) in enumerate(
            zip(n_ctx_q_per_seq, n_ctx_k_per_seq)
        ):
            # Everything below is per sequence
            n_q_ctx_blocks = triton.cdiv(n_ctx_q, block_q)
            n_k_ctx_blocks = triton.cdiv(n_ctx_k, block_k)
            n_pids_in_seq = n_q_ctx_blocks * n_k_ctx_blocks

            for pid in range(n_pids_in_seq):
                q_block_idx = pid // n_k_ctx_blocks
                k_block_idx = pid % n_k_ctx_blocks

                in_q_token_offset = q_block_idx * block_q
                in_k_token_offset = k_block_idx * block_k

                pid_to_out_q_block.append(q_block_idx)
                pid_to_out_k_block.append(k_block_idx)
                pid_to_in_q_token_offset.append(in_q_token_offset + n_in_q_token_so_far)
                pid_to_in_k_token_offset.append(in_k_token_offset + n_in_k_token_so_far)
                pid_to_out_seq_idx.append(seq_idx)

            n_in_q_token_so_far += n_ctx_q
            n_in_k_token_so_far += n_ctx_k

        args = {"dtype": torch.int32, "device": "cuda"}
        return RaggedQkPidLookupTable(
            pid_to_in_q_token_offset=torch.tensor(pid_to_in_q_token_offset, **args),  # type: ignore
            pid_to_in_k_token_offset=torch.tensor(pid_to_in_k_token_offset, **args),  # type: ignore
            pid_to_out_q_block=torch.tensor(pid_to_out_q_block, **args),  # type: ignore
            pid_to_out_k_block=torch.tensor(pid_to_out_k_block, **args),  # type: ignore
            pid_to_out_seq_idx=torch.tensor(pid_to_out_seq_idx, **args),  # type: ignore
            n_pids_total=len(pid_to_in_q_token_offset),
        )


def ragged_single_seq_qk_dotprod(
    query: torch.Tensor, key: torch.Tensor, lut: RaggedQkPidLookupTable
) -> torch.Tensor:
    assert query.ndim == 2 and key.ndim == 2
    device = query.device

    # handle non-contiguous inputs if necessary
    if query.stride(0) > 1 and query.stride(1) > 1:
        query = query.contiguous()
    if key.stride(0) > 1 and key.stride(1) > 1:
        key = key.contiguous()

    # check constraints
    n_ctx_q, d_head = query.shape
    n_ctx_k, d_head_k = key.shape
    assert d_head == d_head_k, f"{query.shape=} {key.shape=}"

    # allocates output
    scores_out = torch.empty((1, n_ctx_q, n_ctx_k), device=device, dtype=query.dtype)

    # Stride along the d_head dimension must be 1
    assert query.stride(1) == 1, f"{query.stride(1)}"
    assert key.stride(1) == 1, f"{key.stride(1)}"

    # pid_to_seq_idx = [0, 0, 1, 2, 2]
    grid = (lut.n_pids_total,)
    _qk_dotprod_kernel[grid](
        q_ptr=query,
        k_ptr=key,
        scores_ptr=scores_out,
        # Lookup tables (sometimes referred to as a "lut")
        pid_to_in_q_token_offset_ptr=lut.pid_to_in_q_token_offset,
        pid_to_in_k_token_offset_ptr=lut.pid_to_in_k_token_offset,
        pid_to_out_q_block_ptr=lut.pid_to_out_q_block,
        pid_to_out_k_block_ptr=lut.pid_to_out_k_block,
        pid_to_out_seq_idx_ptr=lut.pid_to_out_seq_idx,
        # Integers
        max_n_ctx_q_across_seqs=n_ctx_q,
        max_n_ctx_k_across_seqs=n_ctx_k,
        d_head=d_head,
        stride_ctx_q=query.stride(0),
        stride_ctx_k=key.stride(0),
        stride_out_seq=scores_out.stride(0),
        stride_out_q=scores_out.stride(1),
        stride_out_k=scores_out.stride(2),
        total_ctx_q_across_all_seqs=n_ctx_q,
        total_ctx_k_across_all_seqs=n_ctx_k,
    )
    return scores_out.reshape((n_ctx_q, n_ctx_k))


def ragged_qk_dotprod(
    query: RaggedActivations, key: RaggedActivations, lut: RaggedQkPidLookupTable
) -> torch.Tensor:
    device = query.device

    assert query.raw_tensor.is_contiguous()
    assert key.raw_tensor.is_contiguous()

    # check constraints
    total_ctx_q_across_all_seqs, d_head = query.raw_tensor.shape
    total_ctx_k_across_all_seqs, d_head_k = key.raw_tensor.shape
    assert d_head == d_head_k, f"{query.raw_tensor.shape=} {key.raw_tensor.shape=}"

    # allocates output
    # max_n_ctx_q_across_seqs = query.max_n_ctx_per_seq

    assert query.n_seqs == key.n_seqs
    # TODO: flag use zeros for garbage
    scores_out = torch.ones(
        (query.n_seqs, query.max_n_ctx_per_seq, key.max_n_ctx_per_seq),
        device=device,
        dtype=query.dtype,
    )

    # Stride along the d_head dimension must be 1
    assert query.raw_tensor.stride(1) == 1, f"{query.raw_tensor.stride(1)}"
    assert key.raw_tensor.stride(1) == 1, f"{key.raw_tensor.stride(1)}"

    # pid_to_seq_idx = [0, 0, 1, 2, 2]
    grid = (lut.n_pids_total,)
    _qk_dotprod_kernel[grid](
        q_ptr=query.raw_tensor,
        k_ptr=key.raw_tensor,
        scores_ptr=scores_out,
        # Lookup tables (sometimes referred to as a "lut")
        pid_to_in_q_token_offset_ptr=lut.pid_to_in_q_token_offset,
        pid_to_in_k_token_offset_ptr=lut.pid_to_in_k_token_offset,
        pid_to_out_q_block_ptr=lut.pid_to_out_q_block,
        pid_to_out_k_block_ptr=lut.pid_to_out_k_block,
        pid_to_out_seq_idx_ptr=lut.pid_to_out_seq_idx,
        # Integers
        max_n_ctx_q_across_seqs=query.max_n_ctx_per_seq,
        max_n_ctx_k_across_seqs=key.max_n_ctx_per_seq,
        d_head=d_head,
        stride_ctx_q=query.raw_tensor.stride(0),
        stride_ctx_k=key.raw_tensor.stride(0),
        stride_out_seq=scores_out.stride(0),
        stride_out_q=scores_out.stride(1),
        stride_out_k=scores_out.stride(2),
        total_ctx_q_across_all_seqs=total_ctx_q_across_all_seqs,
        total_ctx_k_across_all_seqs=total_ctx_k_across_all_seqs,
    )
    return scores_out



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


"""
A script to run multinode training with submitit.
Almost copy-paste from https://github.com/facebookresearch/deit/blob/main/run_with_submitit.py
"""

import argparse
import os
import uuid
from pathlib import Path

import submitit

from xformers.benchmarks.LRA.run_tasks import benchmark, get_arg_parser


def parse_args():
    parser = argparse.ArgumentParser(
        "Submitit for LRA", parents=[get_arg_parser()], add_help=False
    )
    parser.add_argument(
        "--ngpus", default=1, type=int, help="Number of gpus to request on each node"
    )
    parser.add_argument(
        "--nodes", default=1, type=int, help="Number of nodes to request"
    )
    parser.add_argument("--timeout", default=2800, type=int, help="Duration of the job")

    parser.add_argument(
        "--partition", default="a100", type=str, help="Partition where to submit"
    )
    parser.add_argument(
        "--use_volta32", action="store_true", help="Big models? Use this"
    )
    parser.add_argument(
        "--enforce_host_memory", action="store_true", help="Use if the host OOMs"
    )

    parser.add_argument(
        "--comment",
        default="",
        type=str,
        help="Comment to pass to scheduler, e.g. priority message",
    )
    return parser.parse_args()


def get_shared_folder() -> Path:
    user = os.getenv("USER")
    checkpoint_paths = ["/checkpoint", "/checkpoints"]
    for checkpoint_path in checkpoint_paths:
        if Path(checkpoint_path).is_dir():
            p = Path(f"{checkpoint_path}/{user}/xformers/submitit")
            p.mkdir(exist_ok=True, parents=True)
            return p
    raise RuntimeError(f"No shared folder available - considering {checkpoint_paths}")


def get_init_file():
    # Init file must not exist, but it's parent dir must exist.
    os.makedirs(str(get_shared_folder()), exist_ok=True)
    init_file = get_shared_folder() / f"{uuid.uuid4().hex}_init"
    if init_file.exists():
        os.remove(str(init_file))
    return init_file


class Trainer(object):
    def __init__(self, args):
        self.args = args

    def __call__(self):
        self._setup_gpu_args()
        benchmark(self.args.rank, self.args)

    def checkpoint(self):
        self.args.dist_url = get_init_file().as_uri()
        print("Requeuing ", self.args)
        empty_trainer = type(self)(self.args)
        return submitit.helpers.DelayedSubmission(empty_trainer)

    def _setup_gpu_args(self):
        job_env = submitit.JobEnvironment()
        self.args.checkpoint_dir = Path(
            str(self.args.checkpoint_dir).replace("%j", str(job_env.job_id))
        )
        self.args.gpu = job_env.local_rank
        self.args.rank = job_env.global_rank
        self.args.world_size = job_env.num_tasks
        print(f"Process group: {job_env.num_tasks} tasks, rank: {job_env.global_rank}")


def main():
    args = parse_args()
    if args.checkpoint_dir == "":
        args.checkpoint_dir = get_shared_folder() / "%j"
    Path(args.checkpoint_dir).mkdir(parents=True, exist_ok=True)
    executor = submitit.AutoExecutor(
        folder=args.checkpoint_dir, slurm_max_num_timeout=30
    )

    num_gpus_per_node = args.ngpus
    nodes = args.nodes
    timeout_min = args.timeout
    args.world_size = args.nodes * args.ngpus

    partition = args.partition

    kwargs = {
        "gpus_per_node": num_gpus_per_node,
        "tasks_per_node": num_gpus_per_node,  # one task per GPU
        "cpus_per_task": 10,
        "nodes": nodes,
        "timeout_min": timeout_min,  # max is 60 * 72
        # Below are cluster dependent parameters
        "slurm_partition": partition,
        "slurm_signal_delay_s": 120,
    }

    if args.enforce_host_memory:
        kwargs["mem_gb"] = (40 * num_gpus_per_node,)

    if args.use_volta32:
        kwargs["slurm_constraint"] = "volta32gb"

    if args.comment:
        kwargs["slurm_comment"] = args.comment

    executor.update_parameters(
        **kwargs,
    )

    executor.update_parameters(name="lra")

    args.dist_url = get_init_file().as_uri()
    args.temp_file = str(get_init_file())

    trainer = Trainer(args)
    job = executor.submit(trainer)

    print(f"Submitted job_id: {job.job_id}")
    print(f"Logs and checkpoints will be saved at: {args.checkpoint_dir}")
    with open(Path(f"{args.checkpoint_dir}") / Path("jobs.txt"), "a") as jobfile:
        jobfile.write(f"{job.job_id}\n")


if __name__ == "__main__":
    main()



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


from dataclasses import dataclass
from typing import Optional

import torch
import torch.nn as nn

from xformers.components.positional_embedding import (
    PositionEmbedding,
    PositionEmbeddingConfig,
    register_positional_embedding,
)


@dataclass
class VocabEmbeddingConfig(PositionEmbeddingConfig):
    vocab_size: int
    dropout: float


@register_positional_embedding("vocab", VocabEmbeddingConfig)
class VocabEmbedding(PositionEmbedding):
    def __init__(
        self,
        dim_model: int,
        seq_len: int,
        vocab_size: int,
        dropout: float = 0.0,
        *args,
        **kwargs
    ):
        super().__init__()

        self.vocab_size = vocab_size
        self.dim_model = dim_model

        self.dropout = torch.nn.Dropout(p=dropout)

        self.position_embeddings = nn.Embedding(seq_len, self.dim_model)
        torch.nn.init.normal_(self.position_embeddings.weight, std=0.02)

        self.position_ids: Optional[torch.Tensor] = None

        self.word_embeddings = nn.Embedding(self.vocab_size, self.dim_model)
        torch.nn.init.normal_(self.word_embeddings.weight, std=0.02)

    def forward(self, x: torch.Tensor):
        position_ids = torch.arange(x.shape[1], dtype=torch.long, device=x.device)[
            None, :
        ].repeat(x.shape[0], 1)

        X_token = self.word_embeddings(x)
        X_pos = self.position_embeddings(position_ids)

        X = X_token + X_pos
        X = self.dropout(X)

        return X



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


# CREDITS: adapted from the Nystromformer repo
# https://github.com/mlpen/Nystromformer

from enum import Enum

import torch
import torch.nn as nn

from xformers.components import build_attention
from xformers.components.multi_head_dispatch import MultiHeadDispatchConfig
from xformers.factory import xFormer, xFormerConfig, xFormerEncoderConfig
from xformers.utils import generate_matching_config


class Pooling(str, Enum):
    MEAN = "mean"
    CLS = "cls"


def pooling(mode: Pooling):
    def pool_cls(inp):
        return inp[:, 0, :]

    def pool_mean(inp):
        return inp.mean(dim=1)

    return {Pooling.MEAN: pool_mean, Pooling.CLS: pool_cls}[mode]


def append_cls(inp, mask, vocab_size):
    batch_size = inp.size(0)
    cls_id = (
        (vocab_size - 1) * torch.ones(batch_size, dtype=torch.long, device=inp.device)
    ).long()
    cls_mask = torch.ones(batch_size, dtype=torch.float, device=mask.device)
    inp = torch.cat([cls_id[:, None], inp[:, :-1]], dim=-1)
    mask = torch.cat([cls_mask[:, None], mask[:, :-1]], dim=-1)
    return inp, mask


def patch_model_config(config, attention_name):
    # Rebuild a specific config out of generic + extra params
    commons = config["common"]
    try:
        extra_attention_settings = config["extra_settings"]["attention"][attention_name]
    except KeyError:
        extra_attention_settings = None

    for bc in config["xformer"]:
        bc["dim_model"] = commons["dim_model"]
        bc["position_encoding_config"].update(commons)
        bc["feedforward_config"].update(commons)
        bc["multi_head_config"].update(commons)
        bc["multi_head_config"]["attention"].update(commons)
        bc["multi_head_config"]["attention"]["name"] = attention_name
        bc["multi_head_config"]["attention"]["dim_head"] = (
            commons["dim_model"] / commons["num_heads"]
        )
        if extra_attention_settings is not None:
            bc["multi_head_config"]["attention"].update(extra_attention_settings)

        bc["multi_head_config"] = generate_matching_config(
            bc["multi_head_config"], MultiHeadDispatchConfig
        )
        bc["multi_head_config"].attention = build_attention(
            bc["multi_head_config"].attention
        )
        bc = generate_matching_config(bc, xFormerEncoderConfig)

    return config


class SCHead(nn.Module):
    def __init__(self, config, dim_embedding, dim_mlp):
        super().__init__()
        self.pooling = pooling(Pooling(config["pooling_mode"]))

        self.mlpblock = nn.Sequential(
            nn.Linear(dim_embedding, dim_mlp),
            nn.ReLU(),
            nn.Linear(dim_mlp, config["common"]["num_classes"]),
        )

    def forward(self, inp: torch.Tensor):
        seq_score = self.mlpblock(self.pooling(inp))
        return seq_score


class SCHeadDual(nn.Module):
    def __init__(self, config, dim_embedding, dim_mlp):
        super().__init__()
        self.pooling = pooling(Pooling(config["pooling_mode"]))

        self.mlpblock = nn.Sequential(
            nn.Linear(
                dim_embedding * 4,
                dim_mlp,
            ),
            nn.ReLU(),
            nn.Linear(dim_mlp, config["common"]["num_classes"]),
        )

    def forward(self, inp_0: torch.Tensor, inp_1: torch.Tensor):
        X_0 = self.pooling(inp_0)
        X_1 = self.pooling(inp_1)
        seq_score = self.mlpblock(torch.cat([X_0, X_1, X_0 * X_1, X_0 - X_1], dim=-1))
        return seq_score


class ModelTrunk(nn.Module):
    def __init__(self, config, model_name):
        super().__init__()

        config_model = config["model"]

        self.enable_amp = config["training"]["mixed_precision"]
        self.pooling_mode = Pooling(config_model["pooling_mode"])
        self.vocab_size = config_model["common"]["vocab_size"]

        # Rebuild a specific config out of generic + extra params
        self.config_model = patch_model_config(config_model, model_name)
        self.model = xFormer.from_config(xFormerConfig(config_model["xformer"]))
        self.norm = nn.LayerNorm(self.config_model["common"]["dim_model"])

        ff_config = self.config_model["xformer"][0]["feedforward_config"]
        self.dim_mlp = (
            self.config_model["common"]["dim_model"]
            * ff_config["hidden_layer_multiplier"]
        )


class ModelForSC(ModelTrunk):
    def __init__(self, config, model_name):
        # Setup trunk
        super().__init__(config, model_name)

        self.seq_classifer = SCHead(
            self.config_model,
            dim_embedding=self.config_model["common"]["dim_model"],
            dim_mlp=self.dim_mlp,
        )

    def forward(self, input_ids_0, mask_0, label):

        with torch.cuda.amp.autocast(enabled=self.enable_amp):
            if self.pooling_mode == Pooling.CLS:
                input_ids_0, mask_0 = append_cls(input_ids_0, mask_0, self.vocab_size)

            token_out = self.norm(
                self.model(input_ids_0, encoder_input_mask=mask_0)
            ) * mask_0.unsqueeze(-1)

            seq_scores = self.seq_classifer(token_out)

            seq_loss = torch.nn.CrossEntropyLoss(reduction="none")(seq_scores, label)
            seq_accu = (seq_scores.argmax(dim=-1) == label).to(torch.float32)
            outputs = {
                "loss": seq_loss.mean(),
                "accu": seq_accu.mean(),
                "count": label.size(0),
            }

        return outputs


class ModelForSCDual(ModelTrunk):
    def __init__(self, config, model_name):
        # Setup trunk
        super().__init__(config, model_name)

        self.seq_classifer = SCHeadDual(
            self.config_model,
            dim_embedding=self.config_model["common"]["dim_model"],
            dim_mlp=self.dim_mlp,
        )

    def forward(self, input_ids_0, input_ids_1, mask_0, mask_1, label):

        with torch.cuda.amp.autocast(enabled=self.enable_amp):
            mask_0, mask_1 = mask_0.long(), mask_1.long()

            if self.pooling_mode == Pooling.CLS:
                input_ids_0, mask_0 = append_cls(input_ids_0, mask_0, self.vocab_size)
                input_ids_1, mask_1 = append_cls(input_ids_1, mask_1, self.vocab_size)

            # Concatenate the two inputs into one batch
            input_ids = torch.cat([input_ids_0, input_ids_1], dim=0)
            masks = torch.cat([mask_0, mask_1], dim=0)

            tokens_out = self.norm(
                self.model(input_ids, encoder_input_mask=masks)
            ) * masks.unsqueeze(-1)

            seq_scores = self.seq_classifer(*torch.chunk(tokens_out, 2, dim=0))

            seq_loss = torch.nn.CrossEntropyLoss(reduction="none")(seq_scores, label)
            seq_accu = (seq_scores.argmax(dim=-1) == label).to(torch.float32)
            outputs = {
                "loss": seq_loss.mean(),
                "accu": seq_accu.mean(),
                "count": label.size(0),
            }

        return outputs



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


# CREDITS:
# inspired by
# https://github.com/nateraw/lightning-vision-transformer
# which in turn references https://github.com/lucidrains/vit-pytorch

# Orignal author: Sean Naren

import argparse
import math
from enum import Enum
from typing import Optional

import pytorch_lightning as pl
import torch
import torchvision
from pytorch_lightning.profiler import AdvancedProfiler, PyTorchProfiler
from timm.models.vision_transformer import VisionTransformer as TimmVisionTransformer
from torch import nn
from torch.utils.data import DataLoader, random_split
from torchmetrics import Accuracy
from torchvision import transforms

from xformers.factory import xFormer, xFormerConfig


class Classifier(str, Enum):
    GAP = "gap"
    TOKEN = "token"


class FakeDataModule(pl.LightningDataModule):
    def __init__(
        self,
        batch_size: int,
        pic_size: int = 224,
        num_samples: int = 1000,
        num_classes=10,
        num_workers: int = 2,
        pin_memory: bool = True,
    ):
        super().__init__()

        self.transform = torchvision.transforms.Compose(
            [transforms.ToTensor(), transforms.Normalize((0.5,), (0.5,))]
        )
        self.dataset = torchvision.datasets.FakeData(
            size=num_samples,
            image_size=(3, pic_size, pic_size),
            num_classes=num_classes,
            transform=self.transform,
        )
        self.num_samples = num_samples
        self.pic_size = pic_size
        self.num_classes = num_classes
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.pin_memory = pin_memory

    def setup(self, stage: Optional[str] = None):
        if stage == "fit" or stage is None:
            train_samples = int(self.num_samples * 0.9)
            val_samples = self.num_samples - train_samples

            self.data_train, self.data_val = random_split(
                self.dataset, [train_samples, val_samples]
            )

        # Assign test dataset for use in dataloader(s)
        if stage == "test" or stage is None:
            self.data_test = torchvision.datasets.FakeData(
                size=int(self.num_samples * 0.1),
                image_size=(3, self.pic_size, self.pic_size),
                num_classes=self.num_classes,
                transform=self.transform,
            )

    def train_dataloader(self):
        return DataLoader(
            self.data_train,
            batch_size=32,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
        )

    def val_dataloader(self):
        return DataLoader(
            self.data_val,
            batch_size=32,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
        )

    def test_dataloader(self):
        return DataLoader(
            self.data_test,
            batch_size=32,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
        )


class VisionTransformer(pl.LightningModule):
    def __init__(
        self,
        steps,
        learning_rate=1e-2,
        weight_decay=0.0001,
        image_size=32,
        num_classes=10,
        patch_size=4,
        dim=256,
        n_layer=12,
        n_head=8,
        resid_pdrop=0.1,
        attn_pdrop=0.1,
        mlp_pdrop=0.1,
        attention="scaled_dot_product",
        hidden_layer_multiplier=4,
        linear_warmup_ratio=0.05,
        classifier: Classifier = Classifier.GAP,
        use_timm: bool = False,
    ):

        super().__init__()

        # all the inputs are saved under self.hparams (hyperparams)
        self.save_hyperparameters()

        assert image_size % patch_size == 0

        num_patches = (image_size // patch_size) ** 2

        if use_timm:
            self.transformer = TimmVisionTransformer(
                img_size=image_size,
                patch_size=patch_size,
                in_chans=3,
                num_classes=num_classes,
                embed_dim=dim,
                depth=n_layer,
                num_heads=n_head,
                mlp_ratio=hidden_layer_multiplier,
                qkv_bias=True,
                drop_rate=resid_pdrop,
                attn_drop_rate=attn_pdrop,
                drop_path_rate=0,
                # embed_layer (nn.Module): patch embedding layer
                # norm_layer: (nn.Module): normalization layer
                # weight_init: (str): weight init scheme
            )

        else:
            # A list of the encoder or decoder blocks which constitute the Transformer.
            xformer_config = [
                {
                    "block_type": "encoder",
                    "num_layers": n_layer,
                    "dim_model": dim,
                    "layer_norm_style": "pre",
                    "multi_head_config": {
                        "num_heads": n_head,
                        "residual_dropout": resid_pdrop,
                        "attention": {
                            "name": attention,
                            "dropout": attn_pdrop,
                            "causal": False,
                            "seq_len": num_patches,
                        },
                    },
                    "feedforward_config": {
                        "name": "FusedMLP",
                        "dropout": mlp_pdrop,
                        "activation": "gelu",
                        "hidden_layer_multiplier": hidden_layer_multiplier,
                    },
                }
            ]

            config = xFormerConfig(xformer_config)
            self.transformer = xFormer.from_config(config)

            # init positional embedding with 0.02 from BERT
            self.pos_emb = nn.Parameter(
                torch.randn(1, num_patches + (classifier == Classifier.TOKEN), dim)
                * 0.02
            )
            self.patch_emb = nn.Conv2d(
                3, dim, kernel_size=patch_size, stride=patch_size
            )

            if classifier == Classifier.TOKEN:
                self.clf_token = nn.Parameter(torch.zeros(dim))

            self.ln = nn.LayerNorm(dim)
            self.head = nn.Linear(dim, num_classes)

        self.criterion = torch.nn.CrossEntropyLoss()
        self.val_accuracy = Accuracy()

    @staticmethod
    def linear_warmup_cosine_decay(warmup_steps, total_steps):
        """
        Linear warmup for warmup_steps, with cosine annealing to 0 at total_steps
        """

        def fn(step):
            if step < warmup_steps:
                return float(step) / float(max(1, warmup_steps))

            progress = float(step - warmup_steps) / float(
                max(1, total_steps - warmup_steps)
            )
            return 0.5 * (1.0 + math.cos(math.pi * progress))

        return fn

    def configure_optimizers(self):
        optimizer = torch.optim.SGD(
            self.parameters(),
            lr=self.hparams.learning_rate,
            momentum=0.9,
            weight_decay=self.hparams.weight_decay,
        )

        warmup_steps = int(self.hparams.linear_warmup_ratio * self.hparams.steps)

        scheduler = {
            "scheduler": torch.optim.lr_scheduler.LambdaLR(
                optimizer,
                self.linear_warmup_cosine_decay(warmup_steps, self.hparams.steps),
            ),
            "interval": "step",
        }

        return [optimizer], [scheduler]

    def forward(self, x):
        batch, *_ = x.shape  # BCHW

        if self.hparams.use_timm:
            x = self.transformer(x)
        else:
            x = self.patch_emb(x)

            # flatten patches into sequence
            x = x.flatten(2, 3).transpose(1, 2).contiguous()  # B HW C

            if self.hparams.classifier == Classifier.TOKEN:
                # prepend classification token
                clf_token = (
                    torch.ones(1, batch, self.hparams.dim, device=x.device)
                    * self.clf_token
                )
                x = torch.cat([clf_token, x[:-1, :, :]], axis=0)

            # add position embedding
            x += self.pos_emb.expand_as(x)

            x = self.transformer(x)
            x = self.ln(x)

            if self.hparams.classifier == Classifier.TOKEN:
                x = x[:, 0]
            elif self.hparams.classifier == Classifier.GAP:
                x = x.mean(dim=1)  # mean over sequence len

            x = self.head(x)
        return x

    def training_step(self, batch, _):
        x, y = batch
        y_hat = self(x)

        loss = self.criterion(y_hat, y)

        self.logger.log_metrics(
            {
                "train_loss": loss.mean(),
                "learning_rate": self.lr_schedulers().get_last_lr()[0],
            },
            step=trainer.global_step,
        )

        return loss

    def evaluate(self, batch, stage=None):
        x, y = batch
        y_hat = self(x)
        loss = self.criterion(y_hat, y)
        acc = self.val_accuracy(y_hat, y)

        if stage:
            self.log(f"{stage}_loss", loss, prog_bar=True)
            self.log(f"{stage}_acc", acc, prog_bar=True)

    def validation_step(self, batch, batch_idx):
        self.evaluate(batch, "val")

    def test_step(self, batch, batch_idx):
        pass


if __name__ == "__main__":
    # Get the user requests
    parser = argparse.ArgumentParser("Benchmark xformers and timm ViTs")
    parser.add_argument(
        "-timm",
        "--timm",
        help="Use the ViT implementation from Timm",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "-profile",
        "--profile",
        help="Pofile the runtime and memory",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "-chrome_trace",
        "--chrome_trace",
        help="Pofile the runtime and memory using a chrome trace",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "-samples",
        "--samples",
        help="Number of random samples to generate",
        type=int,
        default=128,
    )

    args = parser.parse_args()

    pl.seed_everything(42)
    BATCH_SIZE = 16
    MAX_EPOCHS = 1
    NUM_WORKERS = 4
    GPUS = 1
    IMG_SIZE = 224
    NUM_SAMPLES = args.samples
    NUM_CLASSES = 10
    N_LAYERS = 4

    # We'll use a dummy datamodule here, which already handles dataset/dataloader/sampler
    dm = FakeDataModule(
        pic_size=IMG_SIZE,
        num_samples=NUM_SAMPLES,
        num_classes=NUM_CLASSES,
        batch_size=BATCH_SIZE,
        num_workers=NUM_WORKERS,
        pin_memory=True,
    )

    # compute total number of steps
    batch_size = BATCH_SIZE * GPUS
    steps = dm.num_samples // batch_size * MAX_EPOCHS
    lm = VisionTransformer(
        steps=steps,
        image_size=IMG_SIZE,
        num_classes=NUM_CLASSES,
        attention="scaled_dot_product",
        use_timm=args.timm,
        n_layer=N_LAYERS,
        patch_size=16,
    )
    profiler = (
        PyTorchProfiler(
            filename="timm_trace" if args.timm else "xFormers_trace",
            export_to_chrome=True,
            record_functions={"training_step_and_backward"},
        )
        if args.profile and args.chrome_trace
        else AdvancedProfiler(
            filename="timm_profile" if args.timm else "xFormers_profile",
        )
        if args.profile
        else None
    )

    trainer = pl.Trainer(
        gpus=GPUS,
        max_epochs=MAX_EPOCHS,
        detect_anomaly=True,
        precision=16,
        profiler=profiler,
    )
    trainer.fit(lm, dm)

    # check the training
    trainer.test(lm, datamodule=dm)



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.

# CREDITS: the underlying kernel comes straight from the Triton tutorials
# see https://github.com/openai/triton/blob/master/python/tutorials/05-layer-norm.py

import logging
from typing import Optional

import torch
import torch.nn as nn
import triton

from xformers.triton.k_layer_norm import _LayerNorm

_triton_registered_warnings = False


class FusedLayerNorm(nn.Module):
    """
    Handle a layer normalization, like torch.nn.LayerNorm_.

    This implementation should be measurably faster than the default PyTorch layernorm (as of PyTorch 1.9),
    both for training and inference worloads.

    .. NOTE: Computations under Torch AMP are kept as float32 by default, one can change this to be float16
        by setting the flag `xformers.triton.k_layer_norm._triton_layernorm_fp16_enabled = True`

    .. _torch.nn.LayerNorm: https://pytorch.org/docs/stable/generated/torch.nn.LayerNorm.html

    """

    def __init__(self, normalized_shape, affine=True, eps=1e-05):
        super().__init__()
        if affine:
            self.weight = nn.Parameter(torch.ones(normalized_shape))
            self.bias = nn.Parameter(torch.zeros(normalized_shape))
        else:
            self.weight = self.bias = None
        self.epsilon = eps

    def forward(self, x):
        return layer_norm(x, self.weight, self.bias, self.epsilon)


def layer_norm(
    x: torch.Tensor,
    weight: Optional[torch.Tensor] = None,
    bias: Optional[torch.Tensor] = None,
    eps: float = 1e-05,
) -> torch.Tensor:

    global _triton_registered_warnings

    r"""Applies normalization over a mini batch of inputs"""

    try:
        if (
            not _triton_registered_warnings
            and torch.cuda.is_available()
            and x.is_cuda
            and weight is not None
            and bias is not None
        ):
            return _LayerNorm.apply(x, weight, bias, eps)
    except (triton.code_gen.OutOfResources, RuntimeError) as e:
        # Catch cases where the current GPU does not have enough registers to hold a full tensor line
        # fallback to PyTorch's implementation, which streams the tensor in and out
        _triton_registered_warnings = True
        logging.warning(
            "Triton layernorm kernel register spillover or invalid image caught. "
            "Deactivating this kernel, please file an issue int the xFormers repository"
        )
        logging.warning(e)

    return torch.nn.functional.layer_norm(
        x, [x.shape[-1]], weight=weight, bias=bias, eps=eps
    )



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


# CREDITS: adapted from the Nystromformer repo
# https://github.com/mlpen/Nystromformer

import argparse
import datetime
import json
import logging
import math
import os
import random
import sys
import time
from contextlib import suppress
from enum import Enum
from pathlib import Path
from typing import Any, Dict

import numpy as np
import torch
import torch.distributed as dist
import torch.nn as nn
from fvcore.nn import FlopCountAnalysis, flop_count_str
from torch.utils.data import DataLoader, DistributedSampler
from torch.utils.tensorboard import SummaryWriter

from xformers.benchmarks.LRA.code.dataset import LRADataset
from xformers.benchmarks.LRA.code.model_wrapper import ModelForSC, ModelForSCDual
from xformers.components.attention import ATTENTION_REGISTRY
from xformers.utils import temp_files_ctx


class Task(str, Enum):
    Retrieval = "retrieval"
    ListOps = "listops"
    Image = "image"
    PathfinderBaseline = "pathfinder32-curv_baseline"
    PathfinderContour9 = "pathfinder32-curv_contour_length_9"
    PathfinderContour14 = "pathfinder32-curv_contour_length_14"
    Text = "text"


def load_config(path: str) -> Dict:
    with open(Path(path).absolute(), "r") as fileio:
        config = json.load(fileio)

    # Duplicate the pathfinder configs
    config["pathfinder32-curv_baseline"] = config["pathfinder32"]
    config["pathfinder32-curv_contour_length_9"] = config["pathfinder32"]
    config["pathfinder32-curv_contour_length_14"] = config["pathfinder32"]
    return config


def build_model(args: argparse.Namespace, config: Dict) -> nn.Module:
    task = args.task
    attention_name = args.attention

    if task == Task.Retrieval:
        model: nn.Module = ModelForSCDual(config[f"{task}"], attention_name)
    else:
        model = ModelForSC(config[f"{task}"], attention_name)

    args.logger.info(model)
    args.logger.info(
        f"num_parameter: {np.sum([np.prod(weight.size()) for weight in model.parameters()]) // 1e3 / 1e3}M"
    )

    with torch.no_grad():
        # Check the flops
        seq_len = config[f"{task}"]["model"]["common"]["seq_len"]
        x = torch.rand(1, seq_len).long()
        mask = torch.rand(1, seq_len).long()
        indices = torch.rand(1, seq_len).long()
        flops = FlopCountAnalysis(model.model, (x, mask, indices))
        args.logger.info(f"complexity: {round(flops.total()/1e9, 3)} GFlops")
        args.logger.info(flop_count_str(flops))

    return model


def build_training_setup(
    config_training: Dict,
    task: Task,
    model: nn.Module,
    rank: int = 0,
    world_size: int = 1,
):
    datasets = {}
    samplers = {}

    for component in ["train", "test", "dev"]:
        dataset = LRADataset(
            file_path=f"datasets/{task}.{component}.pickle",
            seq_len=config_training["seq_len"],
        )

        sampler = DistributedSampler(
            dataset,
            num_replicas=world_size,
            rank=rank,
            shuffle=(component == "train"),
            drop_last=(component == "train"),
        )  # type:ignore
        datasets[component] = dataset
        samplers[component] = sampler

    logging.info(f"Learning rate: {config_training['learning_rate']}")

    optimizer = torch.optim.AdamW(
        model.parameters(),
        lr=config_training["learning_rate"],
        betas=(0.9, 0.999),
        eps=1e-6,
        weight_decay=config_training["weight_decay"],
    )

    lr_scheduler = torch.optim.lr_scheduler.OneCycleLR(  # type: ignore
        optimizer=optimizer,
        max_lr=config_training["learning_rate"],
        pct_start=config_training["warmup"] / config_training["num_train_steps"],
        anneal_strategy=config_training["lr_decay"],
        total_steps=config_training["num_train_steps"],
    )

    amp_scaler = torch.cuda.amp.GradScaler(enabled=config_training["mixed_precision"])

    logging.info(f"Dataloader ready. Rank {rank} of {world_size}")

    return datasets, samplers, optimizer, lr_scheduler, amp_scaler


def print_summary(
    summary,
    save_if_improved,
    train_step_idx,
    model,
    checkpoint_path,
    logger,
    tb_logger=None,
):

    summary["loss"] = np.average(summary["loss"], weights=summary["count"])
    summary["accu"] = np.average(summary["accu"], weights=summary["count"])
    summary["count"] = np.sum(summary["count"]).astype(float)

    if summary["accu"] > summary["best_accu"]:
        summary["best_accu"] = summary["accu"]
        if save_if_improved:
            best_accu = summary["best_accu"]
            torch.save(
                {"model_state_dict": model.state_dict()},
                checkpoint_path,
            )
            logger.info(f"best_accu={best_accu:.3f}. Saved best model")

    summary["max_memory_mb"] = torch.cuda.max_memory_allocated() // 1e3 / 1e3

    summary_round = {"train_step_idx": train_step_idx}
    for key in summary:
        if type(summary[key]) is str:
            summary_round[key] = summary[key]
        else:
            summary_round[key] = round(summary[key], 4)

    if tb_logger:
        tb_logger.add_scalar("acc", summary["accu"], train_step_idx)
        tb_logger.add_scalar("loss", summary["loss"], train_step_idx)
        tb_logger.add_scalar("max_mem", summary["max_memory_mb"], train_step_idx)
        tb_logger.add_scalar("count", summary["count"], train_step_idx)

    logger.info(summary_round)
    logger.info(json.dumps(summary_round, sort_keys=True) + "\n")

    summary["t"] = 0
    summary["loss"] = []
    summary["accu"] = []
    summary["count"] = []


def setup_log(args, rank, attention_name, task):
    log_f = Path(
        os.path.join(
            args.checkpoint_dir, f"{task}__{attention_name}__{rank}_output.log"
        )
    )
    if not log_f.exists():
        log_f.parent.mkdir(parents=True, exist_ok=True)
        with open(log_f, "x") as _:
            pass

    logger = torch.multiprocessing.get_logger()
    logger.setLevel(level=logging.INFO)
    logger.addHandler(logging.FileHandler(filename=str(log_f)))
    if rank == 0:
        logger.addHandler(logging.StreamHandler(sys.stdout))
    return log_f.absolute(), logger


def eval_model(model, dataloaders, component, config, step):
    model.eval()

    for dev_step_idx, batch_dev in enumerate(dataloaders[component]):
        _ = step(
            batch_dev,
            component,
            step_idx=dev_step_idx,
            step_max=config["num_eval_steps"],
        )

        if dev_step_idx == config["num_eval_steps"]:
            break

    model.train()


def rewrite_hyper(config, rewrites):
    def replace(config_dict, k, v):
        if len(k.split(":")) == 1:
            config_dict[k] = v
            return
        first_key = k.split(":")[0]
        assert first_key in config_dict, first_key
        k = k[len(first_key) + 1 :]
        replace(config_dict[first_key], k, v)

    for k, v in rewrites.items():
        replace(config, k, v)
    return config


def seed_worker(_: int):
    # Make sure that non-pytorch random generators are properly set
    worker_seed = torch.initial_seed() % 2 ** 32
    np.random.seed(worker_seed)
    random.seed(worker_seed)


def benchmark(rank, args):
    # Setup multiprocessing
    dist.init_process_group(
        init_method="file://" + args.temp_file,
        backend="NCCL",
        rank=rank,
        world_size=args.world_size,
    )
    try:
        torch.cuda.set_device(args.gpu)
    except AttributeError:
        # Single node launcher
        torch.cuda.set_device(rank)

    task = args.task
    attention_name = args.attention

    # Build the problem
    log_f_path, logger = setup_log(args, rank, attention_name, task)
    args.logger = logger
    config = load_config(args.config)

    config_task = config[f"{task}"]
    if args.sweep_parameters is not None:
        logger.info("Replacing hyperparameters")
        rewrite_hyper(config_task, args.sweep_parameters)

    config_training = config_task["training"]
    config_training["seq_len"] = config_task["model"]["common"]["seq_len"]
    model = build_model(args, config)

    torch.manual_seed(config_training.get("seed", 0))  # also sets the cuda seed
    np.random.seed(config_training.get("seed", 0))
    torch.backends.cudnn.enabled = True
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    torch.cuda.reset_peak_memory_stats()

    # tensorboard
    tb_logger = SummaryWriter(args.tb_dir)

    torch.manual_seed(config_training.get("seed", 0))  # also sets the cuda seed
    np.random.seed(config_training.get("seed", 0))
    torch.backends.cudnn.enabled = True
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    torch.cuda.reset_peak_memory_stats()

    # tensorboard
    tb_logger = SummaryWriter(args.tb_dir)

    # Setup the training
    device_ids = list(range(torch.cuda.device_count()))
    logger.info(f"GPU list: {device_ids}")
    model = model.cuda()
    model = nn.parallel.DistributedDataParallel(
        model, device_ids=[rank], broadcast_buffers=True, find_unused_parameters=True
    )

    (
        datasets,
        samplers,
        optimizer,
        lr_scheduler,
        amp_scaler,
    ) = build_training_setup(config_training, task, model, rank, args.world_size)

    init_t = time.time()

    # Messenger structure which will be moved around to collect metrics
    summary = {
        comp: {
            "t": 0,
            "loss": [],
            "accu": [],
            "count": [],
            "best_accu": 0,
            "component": comp,
        }
        for comp in ["train", "dev", "test"]
    }

    # Setup the dataloaders
    accumu_steps = config_task["training"]["gradient_accumulation"]
    per_gpu_batch_size = (
        config_training["batch_size"] // args.world_size // accumu_steps
    )
    logging.warning(
        "Requested batch size: {}. Given world size and grad accumulation, per-gpu batch is {}".format(
            config_training["batch_size"], per_gpu_batch_size
        )
    )

    # reset train/eval steps if using gradient accumulation
    if accumu_steps > 1:
        config_training["num_train_steps"] *= accumu_steps
        config_training["num_eval_steps"] *= accumu_steps

    epochs = math.ceil(
        config_training["num_train_steps"]
        * config_training["batch_size"]
        / len(datasets["train"])
    )

    logging.warning(
        "Requested train steps: {}. Given dataset, this translates into {} epochs".format(
            config_training["num_train_steps"], epochs
        )
    )

    logger.info(f"accumu_steps={accumu_steps}")
    model_path = str(log_f_path).replace(".log", ".model")
    g = torch.Generator()
    g.manual_seed(config_training.get("seed", 0))

    dataloaders = {
        k: DataLoader(
            datasets[k],
            sampler=samplers[k],
            batch_size=per_gpu_batch_size,
            shuffle=False,
            pin_memory=True,
            num_workers=1,
            worker_init_fn=seed_worker,
            generator=g,
        )
        for k in datasets.keys()
    }

    # Our step function
    def step(
        batch: Dict[str, Any],
        component: str,
        step_idx: int,
        step_max: int,
        accumulate: bool = False,
    ):
        if step_idx > step_max:
            logger.warning(
                "Calling `step` beyond the training schedule, this is probably a mistake"
            )
            return

        t0 = time.time()
        batch_size = batch[list(batch.keys())[0]].size(0)

        for key in batch:
            batch[key] = batch[key].cuda()

        if component == "train":
            acc_context = model.no_sync() if accumulate else suppress()

            with acc_context, torch.autograd.set_detect_anomaly(args.debug):
                outputs = model(**batch)
                amp_scaler.scale(outputs["loss"]).backward()

                if not accumulate:
                    amp_scaler.step(optimizer)
                    optimizer.zero_grad()
                    amp_scaler.update()
                    lr_scheduler.step()

        else:
            with torch.no_grad():
                outputs = model(**batch)

        t1 = time.time()

        t_escape = t1 - t0
        learning_rate = optimizer.param_groups[0]["lr"]
        loss = outputs["loss"].item()
        accu = outputs["accu"].item()
        cnt = outputs["count"]
        time_since_start = time.time() - init_t
        eta = (
            datetime.timedelta(
                seconds=round(time_since_start / (step_idx + 1) * step_max)
            )
            if component == "train"
            else -1
        )

        if not step_idx % 10:
            logger.info(
                f"{component}: step={step_idx}/{step_max}, total_time={time_since_start:.1f},"
                + f" eta={eta},"
                + f" batch_time={t_escape:.3f}, bs={batch_size}, lr={learning_rate:.6f},"
                + f" loss={loss:.4f}, accu={accu:.4f}",
            )

        summary[component]["t"] += t_escape
        summary[component]["loss"].append(loss)
        summary[component]["accu"].append(accu)
        summary[component]["count"].append(cnt)

        if not accumulate:
            step_idx += 1

        return loss, step_idx

    # Start training or evaluating
    train_step_idx = 0
    if not args.skip_train:
        try:
            model.train()
            for epoch in range(epochs):
                logger.info(f"\nEpoch {epoch}")

                # Make sure that per-rank sampling is really random
                for sampler in samplers.values():
                    sampler.set_epoch(epoch)

                for i_batch, batch in enumerate(dataloaders["train"]):
                    grad_accumulate = (
                        i_batch % config_training["gradient_accumulation"] != 0
                    )

                    _, train_step_idx = step(
                        batch,
                        component="train",
                        step_idx=train_step_idx,
                        step_max=config_training["num_train_steps"],
                        accumulate=grad_accumulate,
                    )

                    if not (train_step_idx + 1) % config_training["eval_frequency"]:
                        print_summary(
                            summary["train"],
                            False,
                            train_step_idx,
                            model,
                            model_path,
                            logger,
                        )

                        eval_model(model, dataloaders, "dev", config_training, step)

                        print_summary(
                            summary["dev"],
                            True,
                            train_step_idx,
                            model,
                            model_path,
                            logger,
                            tb_logger,
                        )

                    if train_step_idx == config_training["num_train_steps"]:
                        break

        except KeyboardInterrupt as e:
            print(e)

    checkpoint = torch.load(model_path, map_location="cpu")
    model.load_state_dict(checkpoint["model_state_dict"])
    model.eval()
    try:
        eval_model(model, dataloaders, "test", config_training, step)
    except StopIteration:
        pass

    print_summary(summary["test"], False, train_step_idx, model, model_path, logger)


def get_arg_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--attention",
        type=str,
        help=f"Attention mechanism to chose, among {list(ATTENTION_REGISTRY.keys())}. \
            A list can be passed to test several mechanisms in sequence",
        dest="attention",
        required=True,
    )
    parser.add_argument(
        "--task",
        type=Task,
        help=f"Task to chose, among {[t.value for t in Task]}.",
        dest="task",
        required=True,
    )
    parser.add_argument(
        "--skip_train",
        type=bool,
        help="Whether to skip training, and test an existing model",
        dest="skip_train",
        default=False,
    )
    parser.add_argument(
        "--config",
        type=str,
        help="Path to the config being used",
        dest="config",
        default="./config.json",
    )
    parser.add_argument(
        "--checkpoint_dir",
        type=str,
        help="Path to the checkpoint directory",
        dest="checkpoint_dir",
        default=f"/checkpoints/{os.getenv('USER')}/xformers",
    )
    parser.add_argument(
        "--debug",
        help="Make it easier to debug a possible issue",
        dest="debug",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--world_size",
        help="Number of GPUs used",
        dest="world_size",
        type=int,
        default=1,
    )
    parser.add_argument(
        "--sweep_parameters",
        help="Rewrite some hyperparameters in the config",
        dest="sweep_parameters",
        type=dict,
        default=None,
    )
    parser.add_argument(
        "--tb_dir",
        type=str,
        help="Path to the tensorboard directory",
        dest="tb_dir",
        default=f"/checkpoints/{os.getenv('USER')}/xformers/tb",
    )
    return parser


if __name__ == "__main__":
    parser = get_arg_parser()
    args = parser.parse_args()
    setup_log(args, "main", f"{args.attention}", f"{args.task}")

    with temp_files_ctx(num=1) as temp_files:
        args.temp_file = temp_files[0]
        torch.multiprocessing.spawn(
            benchmark, args=(args,), nprocs=args.world_size, join=True
        )



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


from enum import Enum
from typing import List, Union

import torch
import torch.nn as nn

from xformers import _is_triton_available

if _is_triton_available:
    from xformers.triton.layer_norm import FusedLayerNorm


def _to_tensor_list(
    inputs: Union[torch.Tensor, List[torch.Tensor]]
) -> List[torch.Tensor]:
    if not isinstance(inputs, list):
        inputs = [inputs]
    return inputs


class LayerNormStyle(str, Enum):
    """Support different layer norm styles.
    See "On Layer Normalization in the Transformer Architecture",
    Xiong et al., https://arxiv.org/pdf/2002.04745v1.pdf
    """

    Pre = "pre"
    Post = "post"


# CREDITS: the following is inspired by FastAI's Transformer implementation
class Residual(nn.Module):
    """Object-oriented handling of the residual path"""

    def __init__(self, layer: nn.Module):
        super().__init__()
        self.layer = layer

    def forward(self, inputs: Union[torch.Tensor, List[torch.Tensor]], *args, **kwargs):
        inputs = _to_tensor_list(inputs)

        return inputs[0] + self.layer(*inputs, *args, **kwargs)


class PreNorm(nn.Module):
    """Adds LayerNorm before computing attention

    ..Note: If a list of inputs is passed, all of them get normalized"""

    def __init__(self, d_model: int, sublayer: nn.Module, use_triton: bool = True):
        super().__init__()
        if _is_triton_available and use_triton:
            self.norm: Union[nn.LayerNorm, FusedLayerNorm] = FusedLayerNorm(d_model)
        else:
            self.norm = nn.LayerNorm(d_model)

        self.sublayer = sublayer

    def forward(self, inputs: Union[torch.Tensor, List[torch.Tensor]], *args, **kwargs):
        inputs = _to_tensor_list(inputs)

        x_norm = [self.norm(x_) for x_ in inputs]
        return self.sublayer(*x_norm, *args, **kwargs)


class PostNorm(nn.Module):
    """Adds LayerNorm after computing attention"""

    def __init__(self, d_model: int, sublayer: nn.Module, use_triton: bool = True):
        super().__init__()
        if _is_triton_available and use_triton:
            self.norm: Union[nn.LayerNorm, FusedLayerNorm] = FusedLayerNorm(d_model)
        else:
            self.norm = nn.LayerNorm(d_model)

        self.sublayer = sublayer

    def forward(self, inputs: Union[torch.Tensor, List[torch.Tensor]], *args, **kwargs):
        inputs = _to_tensor_list(inputs)

        x = self.sublayer(*inputs, *args, **kwargs)
        return self.norm(x)



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.

import math
from typing import Any, Optional

import torch
import torch.nn as nn
from torch.cuda.amp import custom_bwd, custom_fwd

from xformers.components.activations import Activation
from xformers.triton.k_activations import (
    get_triton_activation_bwd_kernel,
    get_triton_activation_kernel,
)
from xformers.triton.k_fused_matmul_bw import fused_matmul_backward
from xformers.triton.k_fused_matmul_fw import fused_matmul


class _fused_linear_triton(torch.autograd.Function):
    @staticmethod
    @custom_fwd(cast_inputs=torch.float16)
    def forward(
        ctx,
        x,
        weight,
        bias,
        activation,
        act_grad_kernel,
        trainable_weight,
        trainable_bias,
        save_activation_inputs,
    ):

        # Kick the fused Triton kernel, handling bias and activation in one go
        y, activation_inputs = fused_matmul(
            x, weight, bias, activation, save_activation_inputs
        )

        ctx.activation_grad_kernel = act_grad_kernel
        ctx.trainable_weight = trainable_weight
        ctx.trainable_bias = trainable_bias

        # Micro-optimization: saving these is not always needed (?)
        if x.requires_grad or ctx.trainable_weight or ctx.trainable_bias:
            ctx.save_for_backward(weight, activation_inputs, x)

        return y

    @staticmethod
    @custom_bwd
    def backward(ctx: Any, grad_out: torch.Tensor) -> Any:  # type: ignore
        """
        Compute the derivative with respect to x, other tensors were not trainable inputs.
        """
        (weight, activation_inputs, x) = ctx.saved_tensors

        grad_input, grad_weight, grad_bias = fused_matmul_backward(
            grad_out=grad_out,
            inputs=x,
            act_in=activation_inputs,
            weight=weight,
            trainable_weight=ctx.trainable_weight,
            trainable_bias=ctx.trainable_bias,
            activation_grad=ctx.activation_grad_kernel,
        )

        return grad_input, grad_weight, grad_bias, None, None, None, None, None, None


class FusedLinear(nn.Module):
    """
    Handle a linear transform, like torch.nn.Linear_, and a given activation, in a single kernel.
    The whole transform: is :math:`y = activation(xA^T + b)`.

    This is typically significantly faster than PyTorch while using fp16 and non-sigmoid activations,
    as of September 2021.

    .. _torch.nn.Linear: https://pytorch.org/docs/stable/generated/torch.nn.Linear.html
    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool = False,
        activation: Optional[Activation] = None,
        **_,
    ):
        super().__init__()
        self.weight = nn.Parameter(
            torch.empty(out_features, in_features), requires_grad=True
        )
        self.bias = (
            nn.Parameter(torch.empty(out_features), requires_grad=True)
            if bias
            else None
        )

        self._activation_kernel = get_triton_activation_kernel(activation)
        self._activation_grad_kernel = get_triton_activation_bwd_kernel(activation)

        self.reset_parameters()

    def reset_parameters(self) -> None:
        torch.nn.init.kaiming_uniform_(self.weight, a=math.sqrt(5))
        if self.bias is not None:
            fan_in, _ = torch.nn.init._calculate_fan_in_and_fan_out(self.weight)
            bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            torch.nn.init.uniform_(self.bias, -bound, bound)

    def forward(self, x):
        return _fused_linear_triton.apply(
            x,
            self.weight,
            self.bias,
            self._activation_kernel,
            self._activation_grad_kernel,
            self.weight.requires_grad,
            self.bias.requires_grad if self.bias is not None else False,
            self.training and x.requires_grad,
        )



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


from typing import List

import numpy as np
import torch
import triton
import triton.language as tl


@triton.jit
def garbage_pad_ragged_acts_kernel(
    ragged_acts_ptr,
    ragged_acts_offset_per_seq_ptr,
    n_ctx_per_seq_ptr,
    padded_acts_ptr,
    **meta,  # Optional meta-parameters for the kernel
):
    BLOCK_SIZE = meta["d_model"]  # How many inputs each program should process
    # There are multiple 'program's processing different data. We identify which program
    # we are here

    seq_idx = tl.program_id(axis=0)
    ctx_idx = tl.program_id(axis=1)

    # This program will process inputs that are offset from the initial data.
    # for instance, if you had a vector of length 256 and block_size of 64, the programs
    # would each access the elements [0:64, 64:128, 128:192, 192:256].

    ragged_acts_offset_ptr = ragged_acts_offset_per_seq_ptr + seq_idx
    ragged_acts_offset = tl.load(ragged_acts_offset_ptr)

    # Create a mask to guard memory operations against out-of-bounds accesses
    n_ctx_in_this_seq_ptr = n_ctx_per_seq_ptr + seq_idx
    n_ctx_in_this_seq = tl.load(n_ctx_in_this_seq_ptr)
    ctx_idx_too_large_mask = ctx_idx < n_ctx_in_this_seq

    # Note that offsets is a list of pointers
    ragged_acts_offsets = ragged_acts_offset + tl.arange(0, BLOCK_SIZE)

    # Load ragged acts, since we use a BLOCK_SIZE of d_model, the only out of bounds
    # that we can have is if the n_ctx value is too big
    acts = tl.load(ragged_acts_ptr + ragged_acts_offsets, mask=ctx_idx_too_large_mask)

    # Calculate the offsets for the padded acts
    n_ctx_max = meta["n_ctx_max"]
    padded_acts_offset = n_ctx_max * seq_idx * BLOCK_SIZE

    # Write things back, again masking out the sections that would be garbage
    tl.store(padded_acts_ptr + padded_acts_offset, acts, mask=ctx_idx_too_large_mask)


class RaggedActivations:
    def __init__(self, raw_tensor: torch.Tensor, n_ctx_per_seq: List[int]):
        self.raw_tensor = raw_tensor
        self.n_ctx_per_seq = n_ctx_per_seq

    @property
    def n_seqs(self):
        return len(self.n_ctx_per_seq)

    @property
    def max_n_ctx_per_seq(self):
        return max(self.n_ctx_per_seq)

    @property
    def dtype(self):
        return self.raw_tensor.dtype

    @property
    def device(self):
        return self.raw_tensor.device

    @classmethod
    def from_list(cls, tensors: List[torch.Tensor]):
        """Tensors must all be of shape [n_ctx, d_model]."""
        return cls(
            raw_tensor=torch.cat(tensors),
            n_ctx_per_seq=[tensor.shape[0] for tensor in tensors],
        )

    def iter_full_tensors(self):
        idx_so_far = 0
        for n_ctx_in_this_seq in self.n_ctx_per_seq:
            yield self.raw_tensor[idx_so_far : idx_so_far + n_ctx_in_this_seq]
            idx_so_far += n_ctx_in_this_seq

    def to_garbage_padded(self) -> torch.Tensor:
        """
        Create a tensor of shape (n_seqs, n_ctx_max, d_model) where the
        sequences are right-padded with garbage data
        """
        n_seqs = len(self.n_ctx_per_seq)
        n_ctx_max = max(self.n_ctx_per_seq)

        n_dim = self.raw_tensor.shape[-1]
        # TODO: flag use zeros for garbage
        padded_acts = torch.zeros(
            n_seqs, n_ctx_max, n_dim, dtype=self.raw_tensor.dtype, device="cuda"
        )

        idx_so_far = 0
        for seq_idx, n_ctx_in_this_seq in enumerate(self.n_ctx_per_seq):
            this_seq = self.raw_tensor[idx_so_far : idx_so_far + n_ctx_in_this_seq]
            padded_acts[seq_idx, :n_ctx_in_this_seq, :] = this_seq
            idx_so_far += n_ctx_in_this_seq

        return padded_acts

    def triton_to_garbage_padded(self) -> torch.Tensor:
        """
        Create a tensor of shape (n_seqs, n_ctx_max, d_model) where the
        sequences are right-padded with garbage data
        """
        n_seqs = len(self.n_ctx_per_seq)
        n_ctx_max = max(self.n_ctx_per_seq)

        ragged_acts = self.raw_tensor
        d_model = ragged_acts.shape[-1]
        padded_acts = torch.empty(
            n_seqs, n_ctx_max, d_model, dtype=ragged_acts.dtype, device="cuda"
        )

        # We just use one program per n_ctx position for simplicity
        assert d_model >= 128, f"bad {d_model=}"
        assert d_model <= 8 * 1024, f"bad {d_model=}"
        assert d_model % 32 == 0, f"bad {d_model=}"

        # We use numpy here because it's a bit faster
        n_ctx_per_seq = self.n_ctx_per_seq
        ragged_acts_offset_per_seq = get_acts_offset_per_seq(n_ctx_per_seq)

        # The SPMD launch grid denotes the number of kernel instances that run in parallel.
        # It is analogous to CUDA launch grids. It can be either Tuple[int], or
        # Callable(metaparameters) -> Tuple[int]
        #
        # In this case, we use a 2D grid where the size is n_ctx
        grid_2d = (n_seqs, n_ctx_max)

        # NOTE:
        #  - each torch.tensor object is implicitly converted into a pointer to its
        #       first element.
        #  - `triton.jit`'ed functions can be indexed with a launch grid to obtain a
        #       callable GPU kernel

        # [breakpoint()]
        garbage_pad_ragged_acts_kernel[grid_2d](
            ragged_acts,
            torch.tensor(ragged_acts_offset_per_seq, device="cuda"),
            torch.tensor(self.n_ctx_per_seq, device="cuda"),
            padded_acts,
            d_model=d_model,
            n_ctx_max=n_ctx_max,
        )
        return padded_acts


def get_acts_offset_per_seq(n_ctx_per_seq):
    n_ctx_per_seq_shifted = np.array([0] + n_ctx_per_seq[:-1])
    ragged_acts_offset_per_seq = n_ctx_per_seq_shifted.cumsum(axis=0)
    return ragged_acts_offset_per_seq


"""

# TODO: Build LUT
seq_idx = 1
ctx_idx = 0

ragged_offset = 1

# How to do a list of tensors?
#

# TODO: Add the QK dotprod to get scores
#  - Start with a ragged tensor for the keys also
#  - Using a list of tensors as the Keys
#  - Using sequences

# 16x16x256


# scores [n_seq, n_ctx_keys_max, n_ctx_queries_max]


# final_out [n_seq, n_ctx_keys_max, d_model]
"""



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


from typing import Any, Dict

import torch
import triton

from xformers.benchmarks.utils import TestCase, pretty_plot, pretty_print
from xformers.triton import FusedLayerNorm

SHAPES = [
    (8, 256, 512),
    (8, 512, 1024),
    (4, 1024, 1024),
    (2, 2048, 2048),
    (2, 4096, 4096),
    (1, 2048, 12288),
]


def to_gbs_fw(a, ms):
    # Read and write the full array
    return (2 * a.numel() * a.element_size() * 1e-9) / (ms * 1e-3)


def bench_layernorm(backward: bool):
    device = torch.device("cuda")

    for dtype in [
        torch.float16,
        torch.float32,
    ]:
        results: Dict[str, Any] = {}

        for B, M, K in SHAPES:
            a = torch.rand(B, M, K, device=device, dtype=dtype, requires_grad=backward)

            # Pytorch layer norn
            torch_layernorm = torch.nn.LayerNorm([K]).to(dtype=dtype, device=device)

            # pyre-ignore[16]: TODO(T101400990): Pyre did not recognize the
            # `FusedLinearNorm` import.
            # Fused layernorm equivalent
            fused_layernorm = FusedLayerNorm([K]).to(dtype=dtype, device=device)

            def torch_step(x):
                y = torch_layernorm(x)
                if backward:
                    torch.norm(y).backward()
                return y

            def triton_step(x):
                y = fused_layernorm(x)
                if backward:
                    torch.norm(y).backward()
                return y

            for testcase in [
                TestCase(
                    torch_step,
                    "pytorch - fw{}".format("+bw" if backward else ""),
                ),
                TestCase(
                    triton_step,
                    "triton - fw{}".format("+bw" if backward else ""),
                ),
            ]:
                time = triton.testing.do_bench(lambda: testcase.function(a))[0]
                key = f"B={B}, M={M}, K={K}"
                if key not in results:
                    results[key] = {}

                # Record BW
                bandwidth = to_gbs_fw(a, time)
                results[key][testcase.name] = f"{bandwidth:.1f}"

        pretty_print(results, title="\n --- Type: {} --- ".format(dtype), units="GB/s")
        pretty_plot(
            results,
            title="LayerNorm-FW{}-{}".format("+BW" if backward else "", dtype),
            units="GB/s",
            dash_key="pytorch",
        )


for bw in [False, True]:
    bench_layernorm(bw)



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.

import random
import time
from typing import Any, Dict, List, Tuple

import torch
import triton
from torch.cuda.amp import autocast

from xformers.benchmarks.utils import TestCase, pretty_print
from xformers.factory.model_factory import xFormer, xFormerConfig

VOCAB = 8


def _data(device, batch, seq, emb, vocab=VOCAB):
    # The dummy task is basically to classify sequences, either pure zeroes or some noise
    input_a = torch.zeros((batch, seq, emb), device=device)
    input_b = (torch.rand((batch, seq, emb), device=device) * vocab).abs()

    target_a = torch.zeros((batch, seq), device=device)
    target_b = torch.ones((batch, seq), device=device)

    if random.random() > 0.5:
        return torch.cat([input_a, input_b], dim=0), torch.cat(
            [target_a, target_b], dim=0
        )

    return torch.cat([input_b, input_a], dim=0), torch.cat([target_b, target_a], dim=0)


def reset_seeds():
    torch.manual_seed(0)
    random.seed(0)


def step(
    model: torch.nn.Module,
    optim: torch.optim.Optimizer,
    batch: int,
    seq: int,
    emb: int,
    device,
):
    model.train()
    optim.zero_grad()
    batch, target = _data(device, batch, seq, emb)

    try:
        outputs = model(batch)
    except TypeError:
        # Pytorch encoder exposes target explicitly
        outputs = model(batch, tgt=batch)

    loss = torch.norm(torch.mean(outputs, dim=-1) - target)
    loss.backward()

    # Clip grad and error out if we're producing NaNs
    torch.nn.utils.clip_grad_norm_(model.parameters(), 10.0, norm_type=2.0, error_if_nonfinite=True)  # type: ignore
    optim.step()

    return loss.item()


def evaluate(model: torch.nn.Module, batch: int, seq: int, emb: int, device):
    reset_seeds()
    batch, target = _data(device, batch, seq, emb)
    model.eval()
    try:
        outputs = model(batch)
    except TypeError:
        # Pytorch decoder exposes target explicitly
        outputs = model(batch, tgt=batch)

    return torch.norm(torch.mean(outputs, dim=-1) - target).item()


def train(model, optimizer, name, steps, batch: int, seq: int, emb: int, device):
    # Dummy training, just checking that both options give the same results
    # Same seed for everyone
    start = time.time()
    for _ in range(steps):
        _ = step(model, optimizer, batch, seq, emb, device)

    torch.cuda.synchronize()
    print("Trained {} in {:.3}s".format(name, time.time() - start))


def bench_pytorch_encoder(
    shapes: List[Tuple[int, int, int]],
    activation: str,
    n_heads: int,
    dropout: float = 0.1,
    layers: int = 2,
    device: torch.device = torch.device("cuda"),
    steps: int = 20,
    use_amp: bool = True,
):
    results_time: Dict[str, Any] = {}
    results_memory: Dict[str, Any] = {}

    for shape in shapes:
        batch, seq, emb = shape

        # Build both a xFormers and Pytorch model
        reset_seeds()

        model_xformers = xFormer.from_config(
            xFormerConfig(
                [
                    {
                        "block_type": "encoder",
                        "dim_model": emb,
                        "num_layers": layers,
                        "layer_norm_style": "post",
                        "multi_head_config": {
                            "num_heads": n_heads,
                            "residual_dropout": dropout,
                            "use_separate_proj_weight": True,
                            "bias": True,
                            "attention": {
                                "name": "scaled_dot_product",
                                "dropout": dropout,
                                "causal": False,
                                "seq_len": seq,
                            },
                            "dim_model": emb,
                        },
                        "feedforward_config": {
                            "name": "FusedMLP",
                            "dropout": dropout,
                            "activation": activation,
                            "hidden_layer_multiplier": 4,
                            "dim_model": emb,
                        },
                    },
                ]
            )
        ).to(device)
        print(model_xformers)

        reset_seeds()
        model_pytorch = torch.nn.TransformerEncoder(
            torch.nn.TransformerEncoderLayer(
                d_model=emb,
                nhead=n_heads,
                dim_feedforward=4 * emb,
                dropout=dropout,
                activation=activation,
                layer_norm_eps=1e-05,
                batch_first=True,  # (batch, seq, feature)
                device=device,
            ),
            num_layers=layers,
        )
        print(model_pytorch)

        optim_xformers = torch.optim.Adam(model_xformers.parameters(), lr=1e-3)
        optim_pytorch = torch.optim.Adam(model_pytorch.parameters(), lr=1e-3)

        def run_training(model, optimizer, label):
            with autocast(enabled=use_amp):
                eval_start = evaluate(model, batch, seq, emb, device)
                torch.cuda.empty_cache()
                torch.cuda.reset_peak_memory_stats()
                torch.cuda.synchronize()

                train(model, optimizer, label, steps, batch, seq, emb, device)
                max_memory = torch.cuda.max_memory_allocated() // 2 ** 20
                print(f"Peak memory use: {max_memory}MB")

                eval_stop = evaluate(model, batch, seq, emb, device)
                print(f"Trained from {eval_start} to {eval_stop}\n")
                return eval_start, eval_stop, max_memory

        # Save the memory being used by both
        memory: Dict[str, Any] = {"pytorch": [], "xformers": []}

        def torch_train():
            _, _, max_memory = run_training(model_pytorch, optim_pytorch, "pytorch")
            memory["pytorch"].append(max_memory)

        def xformers_train():
            _, _, max_memory = run_training(model_xformers, optim_xformers, "xformers")
            memory["xformers"].append(max_memory)

        for testcase in [
            TestCase(
                xformers_train,
                "xformers",
            ),
            TestCase(
                torch_train,
                "pytorch",
            ),
        ]:
            time, _, _ = triton.testing.do_bench(lambda: testcase.function())
            key = "emb {} - heads {}".format(emb, n_heads)
            if key not in results_time:
                results_time[key] = {}
                results_memory[key] = {}

            results_time[key][testcase.name] = f"{time/1000:.1f}"

            median_memory = sorted(memory[testcase.name])[
                len(memory[testcase.name]) // 2
            ]
            results_memory[key][testcase.name] = median_memory

    pretty_print(
        results_time,
        title="\n--- Transformer training benchmark - runtime ---",
        units="s",
    )
    pretty_print(
        results_memory,
        title="\n--- Transformer training benchmark - memory use ---",
        units="MB",
    )


if __name__ == "__main__":
    device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
    bench_pytorch_encoder(
        shapes=[(16, 128, 128), (2, 1024, 1024), (1, 1024, 2048)],
        activation="gelu",
        n_heads=8,
        dropout=0.1,
        layers=2,
        device=device,
        steps=20,
        use_amp=True,
    )



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


from abc import ABCMeta, abstractmethod
from dataclasses import asdict, dataclass
from typing import Optional, Type, TypeVar

import torch.nn as nn

from xformers.components import Activation

Self = TypeVar("Self", bound="Feedforward")


@dataclass
class FeedforwardConfig:
    name: str
    dim_model: int
    dropout: float
    activation: Activation


# Define the common interface, every feedforward block needs to derive from it
class Feedforward(nn.Module, metaclass=ABCMeta):
    @abstractmethod
    def __init__(
        self,
        dim_model: Optional[int] = None,
        dropout: Optional[float] = None,
        activation: Optional[Activation] = None,
        *args,
        **kwargs,
    ):
        super().__init__()
        self.requires_cuda = False

    @classmethod
    def from_config(cls: Type[Self], config: FeedforwardConfig) -> Self:
        # Generate the class inputs from the config
        fields = asdict(config)

        # Skip all Nones so that default values are used
        fields = {k: v for k, v in fields.items() if v is not None}

        return cls(**fields)



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


# CREDITS: This comes almost as-is from the Triton dropout tutorial
# https://raw.githubusercontent.com/openai/triton/master/python/tutorials/04-low-memory-dropout.py

from typing import Optional

import torch
import triton
from torch.cuda.amp import custom_bwd, custom_fwd

from xformers.components.activations import Activation, build_activation
from xformers.triton.k_activations import (
    get_triton_activation_bwd_kernel,
    get_triton_activation_kernel,
)
from xformers.triton.k_dropout import k_dropout_bw, k_dropout_fw

GROUP_M = 32
BLOCK_M = GROUP_M // 4
BLOCK_N = 128


# Helper to handle the SPMD launch grid and error cases
class _dropout(torch.autograd.Function):
    @staticmethod
    @custom_fwd(cast_inputs=torch.float16)
    def forward(ctx, x, p, bias, activation, activation_grad, trainable_bias):
        # Soft-flatten an hypothetical 3rd dimension
        x_ = x.reshape(-1, x.shape[-1]).contiguous()
        y = torch.empty_like(x_)
        M, N = x_.shape

        assert bias is None or (bias.dtype == x.dtype and bias.shape[0] == N)

        def grid(meta):
            # NOTE: We use Triton Philox random number generator, which optimally generates 4 blocks for
            # a given seed and offsets. "BLOCK_M" here describes the size of one of these blocks
            # but we need to take this factor of 4 into account when scheduling all the kernels
            return (
                triton.cdiv(M, meta["BLOCK_M"] * 4),
                triton.cdiv(N, meta["BLOCK_N"]),
            )

        N_BLOCK_N = triton.cdiv(N, BLOCK_N)

        # Generate one seed per sample
        # seed max is int32 max for positive numbers: 2**16
        seeds = torch.randint(65536, (N_BLOCK_N,), device=x.device).to(torch.int32)

        # fmt: off
        k_dropout_fw[grid](
            y, x_,
            bias if bias is not None else x_,
            seeds,
            y.stride(0),
            M, N,
            p,
            x.dtype == torch.float16,
            USE_BIAS=bias is not None,
            ACTIVATION=activation,
            BLOCK_M=BLOCK_M,
            BLOCK_N=BLOCK_N,
        )
        # fmt: on

        if activation is not None:
            ctx.save_for_backward(seeds, bias, x)
        else:
            ctx.save_for_backward(seeds, bias, None)

        ctx.trainable_bias = bias is not None and trainable_bias
        ctx.activation_grad = activation_grad
        ctx.p = p

        return y.reshape_as(x)

    @staticmethod
    @custom_bwd
    def backward(ctx, grad_out):
        (seeds, bias, inputs) = ctx.saved_tensors

        # Soft-flatten an hypothetical 3rd dimension
        grad_out_ = grad_out.reshape(-1, grad_out.shape[-1]).contiguous()
        grad_in = torch.empty_like(grad_out_)

        M, N = grad_out_.shape

        # Optional inputs to compute the activation contribution to the gradient
        assert inputs is not None or ctx.activation_grad is None

        if inputs is None:
            inputs = grad_out_
        elif inputs.ndim > 2:
            inputs = inputs.reshape(-1, N)

        # We split the problem in tiles:
        # - over M there will be a follow up reduction
        # - over M, we go by 4 tiles at at time (consequence of the random number generation)
        # - over N we compromise in between trying to use as much memory paralellism as possible,
        # (fill in the warps, there are 32 threads per warps, and 4 warps default), and not being too
        # big because of register spilling
        N_BLOCKS_M = triton.cdiv(M, GROUP_M)

        if ctx.trainable_bias:
            grad_bias = torch.empty(
                (
                    N_BLOCKS_M,
                    N,
                ),
                device=grad_in.device,
                dtype=grad_in.dtype,
            )

        else:
            grad_bias = grad_in  # will not be used

        def grid(meta):
            # NOTE: We use Triton Philox random number generator, which optimally generates 4 blocks for
            # a given seed and offsets. "BLOCK_M" here describes the size of one of these blocks
            # but we need to take this factor of 4 into account when scheduling all the kernels
            return (
                triton.cdiv(M, meta["BLOCK_M"] * 4),
                triton.cdiv(N, meta["BLOCK_N"]),
            )

        # fmt: off
        k_dropout_bw[grid](
            grad_in, grad_bias, grad_out_,
            inputs, bias if bias is not None else inputs,
            seeds,
            grad_out_.stride(0), inputs.stride(0),
            M, N,
            ctx.p,
            grad_in.dtype == torch.float16,
            USE_BIAS=bias is not None,
            ACTIVATION_GRAD=ctx.activation_grad,
            TRAINABLE_BIAS=ctx.trainable_bias,
            BLOCK_M=BLOCK_M,
            BLOCK_N=BLOCK_N,
        )
        # fmt: on

        return (
            grad_in.reshape_as(grad_out),
            None,
            torch.sum(grad_bias, dim=0) if ctx.trainable_bias else None,
            None,
            None,
            None,
        )


def dropout(
    x: torch.Tensor,
    p: float,
    bias: Optional[torch.Tensor] = None,
    activation: Optional[Activation] = None,
):
    """
    Apply dropout on the input tensor.
    Optionally add a bias, the computation will be fused.
    """

    # Micro optim, skip dropout
    if p == 0.0 and activation is None:
        return x + bias if bias is not None else x

    act_kernel = get_triton_activation_kernel(activation)
    act_grad_kernel = get_triton_activation_bwd_kernel(activation)
    return _dropout.apply(
        x,
        p,
        bias,
        act_kernel,
        act_grad_kernel,
        bias is not None and bias.requires_grad,
    )


class FusedDropoutBias(torch.nn.Module):
    def __init__(
        self,
        p: float,
        bias_shape: Optional[int],
        activation: Optional[Activation] = None,
    ) -> None:
        super().__init__()
        self.p = p
        self.activation_type = activation
        self.bias = (
            torch.zeros(bias_shape, requires_grad=True)
            if bias_shape is not None
            else None
        )
        self.activation = get_triton_activation_kernel(activation)
        self.pytorch_activation = build_activation(self.activation_type)
        self.activation_grad = get_triton_activation_bwd_kernel(activation)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # Convenience, catch a possible type or device mismatch
        if self.bias is not None:  # type: ignore
            self.bias = self.bias.to(dtype=x.dtype, device=x.device)  # type: ignore

        # Train/inference
        p = self.p if self.training else 0.0

        # This kernel is slower than pytorch for small buffers, bypassing it in that case
        perf_check = x.shape[-1] > 512

        # Catch a non-cuda setup, fallback to pytorch
        if not x.is_cuda or not perf_check:
            x = x + self.bias if self.bias is not None else x
            x = self.pytorch_activation(x)
            return torch.nn.functional.dropout(x, p)

        # The normal, Triton-backed path
        return _dropout.apply(
            x, p, self.bias, self.activation, self.activation_grad, True
        )



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.

import logging
from dataclasses import dataclass
from typing import Optional, Union

import torch
from torch import nn

from xformers.components.attention import (
    Attention,
    AttentionConfig,
    AttentionMask,
    register_attention,
)
from xformers.components.attention.core import scaled_dot_product_attention


@dataclass
class ScaledDotProductConfig(AttentionConfig):
    causal: Optional[bool]
    seq_len: Optional[int]
    to_seq_len: Optional[int]


@register_attention("scaled_dot_product", ScaledDotProductConfig)
class ScaledDotProduct(Attention):
    r"""
    Implementing the Scaled Dot-Product attention proposed in
    `Attention is all you need`_, Vaswani et al.

    .. _`Attention is all you need`: https://arxiv.org/abs/1706.03762v5
    """

    mask: Optional[AttentionMask]

    def __init__(
        self,
        dropout: float = 0.0,
        causal: bool = False,
        seq_len: Optional[int] = None,
        to_seq_len: Optional[int] = None,
        *args,
        **kwargs,
    ):
        super().__init__()

        self.attn_drop = nn.Dropout(dropout, inplace=False)
        self.causal = causal
        self.seq_len = seq_len

        if causal and seq_len is not None:
            self.mask = AttentionMask.make_causal(seq_len, to_seq_len)
        else:
            self.mask = None

    def forward(
        self,
        q: torch.Tensor,
        k: torch.Tensor,
        v: torch.Tensor,
        att_mask: Optional[Union[AttentionMask, torch.Tensor]] = None,
        *args,
        **kwargs,
    ) -> torch.Tensor:
        r"""
        att_mask    A 2D or 3D mask which ignores attention at certain positions.

                    - If the mask is boolean, a value of True will keep the value,
                        while a value of False will mask the value.

                        Key padding masks (dimension: batch x sequence length) and attention masks
                        (dimension: sequence length x sequence length OR batch x sequence length x sequence length)
                        can be combined and passed in here. Method maybe_merge_masks provided in the utils can be
                        used for that merging.

                    - If the mask has the float type, then an additive mask is expected (masked values are -inf)

        """

        # Convenience, create an attention mask if a tensor was passed
        if att_mask is not None and isinstance(att_mask, torch.Tensor):
            # By default we don't know of the causality, and a check would be expensive
            att_mask = (
                AttentionMask.from_bool(att_mask)
                if att_mask.dtype == torch.bool
                else AttentionMask(att_mask, is_causal=False)
            )

        # Handle a possibly deferred causal mask handling
        if self.causal and self.mask is None:
            self.mask = AttentionMask.make_causal(
                seq_len=q.shape[-2],
                to_seq_len=q.shape[-2],
                device=q.device,
                dtype=q.dtype,
            )

        # Merge the optional causal mask and the user-provided mask
        if self.mask is not None:
            self.mask = self.mask.to(dtype=q.dtype, device=q.device)

            att_mask = att_mask + self.mask if att_mask is not None else self.mask

        # Try to handle a case where the sequence is smaller than the mask
        if (
            att_mask is not None
            and q.shape[-2] == k.shape[-2]
            and q.shape[-2] < att_mask.shape[1]
        ):
            if isinstance(att_mask, AttentionMask):
                att_mask = att_mask.make_crop(seq_len=q.shape[-2])
            else:
                logging.error(
                    "Mismatching sparse attention mask and sequence length."
                    + " Please pad the inputs or adjust the attention mask"
                )
                raise NotImplementedError

        # Attend: (B x nh, S, hs) x (B x nh, hs, S) -> (B x nh, S, S)
        y = scaled_dot_product_attention(
            q=q, k=k, v=v, att_mask=att_mask, dropout=self.attn_drop
        )
        return y



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.

import logging
from pathlib import Path
from typing import Any, Callable, Dict, Set, Union

import torch

from xformers.utils import (
    generate_matching_config,
    get_registry_decorator,
    import_all_modules,
)

from ._sputnik_sparse import SparseCS
from .attention_mask import AttentionMask
from .base import Attention, AttentionConfig  # noqa

# CREDITS: Classy Vision registry mechanism

ATTENTION_REGISTRY: Dict[str, Any] = {}
ATTENTION_CLASS_NAMES: Set[str] = set()

# Arbitrary threshold for now,
# in between dense and sparse matrix algorithms for the attention mechanism
_DENSITY_THRESHOLD = 0.30  # noqa # from the sputnik paper, vs.
_USE_SPUTNIK = True


def build_attention(config: Union[Dict[str, Any], AttentionConfig]):
    """Builds an attention from a config.

    This assumes a 'name' key in the config which is used to determine what
    attention class to instantiate. For instance, a config `{"name": "my_attention",
    "foo": "bar"}` will find a class that was registered as "my_attention"
    (see :func:`register_attention`) and call .from_config on it."""

    if not isinstance(config, AttentionConfig):
        try:
            config_instance = generate_matching_config(
                config, ATTENTION_REGISTRY[config["name"]].config
            )
        except KeyError as e:
            name = config["name"]
            logging.warning(f"{name} not available among {ATTENTION_REGISTRY.keys()}")
            raise e
    else:
        config_instance = config

    return ATTENTION_REGISTRY[config_instance.name].constructor.from_config(
        config_instance
    )


"""Registers an Attention subclass.

    This decorator allows xFormers to instantiate a subclass of Attention
    from a configuration file, even if the class itself is not part of the
    xFormers library. To use it, apply this decorator to an Attention
    subclass, like this:

    .. code-block:: python

        @dataclass
        class MyConfig:
            ...

        @register_attention('my_attention', MyConfig)
        class MyAttention(Attention):
            ...

    To instantiate an attention from a configuration file, see :func:`build_attention`."""
register_attention: Callable[[str, Any], Callable[[Any], Any]] = get_registry_decorator(
    ATTENTION_REGISTRY, ATTENTION_CLASS_NAMES, Attention, AttentionConfig
)


def maybe_sparsify(matrix) -> Any:
    # Sparsify if that makes sense
    if torch.count_nonzero(matrix).item() / matrix.numel() > _DENSITY_THRESHOLD:
        # If not sparse, then AttentionMask is the reference type
        return AttentionMask.from_bool(matrix)

    return sparsify(matrix)


def sparsify(matrix):
    if _USE_SPUTNIK:
        return SparseCS(matrix)
    return matrix.to_sparse()


from .favor import FavorAttention  # noqa
from .global_tokens import GlobalAttention  # noqa
from .linformer import LinformerAttention  # noqa
from .local import LocalAttention  # noqa
from .nystrom import NystromAttention  # noqa
from .ortho import OrthoFormerAttention  # noqa
from .random import RandomAttention  # noqa
from .scaled_dot_product import ScaledDotProduct  # noqa

__all__ = [
    "ScaledDotProduct",
    "LocalAttention",
    "LinformerAttention",
    "NystromAttention",
    "RandomAttention",
    "OrthoFormerAttention",
    "GlobalAttention",
    "FavorAttention",
    "Attention",
    "AttentionMask",
    "build_attention",
    "register_attention",
]

# Optionally expose the BlockSparse attention
try:
    from .blocksparse import BlockSparseAttention  # noqa

    __all__ += ["BlockSparseAttention"]
except ImportError:
    pass


# automatically import any Python files in the directory
import_all_modules(str(Path(__file__).parent), "xformers.components.attention")



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


import math
from dataclasses import dataclass
from typing import Optional, Tuple

import torch
from torch import nn
from torch.nn.init import (
    _calculate_fan_in_and_fan_out,
    _no_grad_uniform_,
    constant_,
    xavier_uniform_,
)


def small_init_(tensor: torch.Tensor, gain: float = 1.0) -> torch.Tensor:
    r"""Fills the input `Tensor` with values according to the method
    described in `Transformer Without Tears`_, using a uniform distribution.

    This is a variation of the Xavier init. The resulting tensor will have values sampled from
    :math:`\mathcal{U}(-a, a)` where

    .. math::
        a = \text{gain} \times \sqrt{\frac{6}{\text{fan\_in} + 4 * \text{fan\_out}}}

    Also known as Glorot initialization.

    Args:
        tensor: an n-dimensional `torch.Tensor`
        gain: an optional scaling factor

    .. _`Transformer Without Tears`: https://doi.org/10.5281/zenodo.3525484

    """
    fan_in, fan_out = _calculate_fan_in_and_fan_out(tensor)
    std = gain * math.sqrt(2.0 / float(fan_in + 4 * fan_out))
    a = math.sqrt(3.0) * std  # Calculate uniform bounds from standard deviation

    return _no_grad_uniform_(tensor, -a, a)


@dataclass
class InProjParams:
    in_features: int
    out_features: int
    bias: bool
    small_init: bool = False


class InProjContainer(nn.Module):
    """
    Handle all the input projections in one go, opportunistically fuse some operations.

    CREDITS: Inspired by https://github.com/pytorch/text/blob/master/torchtext/nn/modules/multiheadattention.py
    and the MultiHeadAttention implementation from PyTorch
    """

    def __init__(
        self,
        query_proj_params: InProjParams,
        key_proj_params: Optional[InProjParams],
        value_proj_params: Optional[InProjParams],
    ):

        super().__init__()

        assert (
            query_proj_params.in_features == query_proj_params.out_features
        ), "We assume in_features == out_features for queries, please provide your projection if this is not the case"

        # If nothing is specified for key and value, use the same as query
        if key_proj_params is None:
            key_proj_params = query_proj_params

        if value_proj_params is None:
            value_proj_params = query_proj_params

        # Catch a beneficial case, if Q,K,V dimensions are the same
        self.same_dimensions = (
            query_proj_params.in_features == key_proj_params.in_features
            and value_proj_params.in_features == key_proj_params.in_features
        )

        self.out_features = query_proj_params.out_features
        self.q_p_params = query_proj_params
        self.k_p_params = key_proj_params
        self.v_p_params = value_proj_params

        # - handle all the weights
        # save the requested init method
        if self.same_dimensions:
            # We can use a single weight and bias buffer, which will speed up self attention
            self.in_proj_weight = nn.Parameter(
                torch.empty((3 * self.out_features, self.out_features))
            )
            self.register_parameter("q_proj_weight", None)
            self.register_parameter("k_proj_weight", None)
            self.register_parameter("v_proj_weight", None)
        else:
            # The dimensions are different, use seperate buffers
            self.q_proj_weight = nn.Parameter(
                torch.empty((self.out_features, query_proj_params.in_features))
            )
            self.k_proj_weight = nn.Parameter(
                torch.empty((self.out_features, key_proj_params.in_features))
            )
            self.v_proj_weight = nn.Parameter(
                torch.empty((self.out_features, value_proj_params.in_features))
            )
            self.register_parameter("in_proj_weight", None)

        # - handle all the inputs
        if query_proj_params.bias:
            self.in_proj_bias = nn.Parameter(torch.empty(3 * self.out_features))
        else:
            self.register_parameter("in_proj_bias", None)

        # - multi-head attention specific init for the weights and biases
        self._reset_parameters()

    def _reset_parameters(self):
        if self.in_proj_weight is not None:
            # 1/sqrt(2) init is empirically beneficial in that case
            self.in_proj_weight = self._init_weights(
                self.q_p_params, self.in_proj_weight, gain=1.0 / math.sqrt(2)
            )

        else:
            self.q_proj_weight = self._init_weights(
                self.q_p_params, self.q_proj_weight, gain=1.0
            )
            self.k_proj_weight = self._init_weights(
                self.k_p_params, self.k_proj_weight, gain=1.0
            )
            self.v_proj_weight = self._init_weights(
                self.v_p_params, self.v_proj_weight, gain=1.0
            )

        if self.in_proj_bias is not None:
            constant_(self.in_proj_bias, 0.0)

    @staticmethod
    def _init_weights(params: InProjParams, weights: torch.Tensor, gain: float):
        if params.small_init:
            return small_init_(weights, gain=gain)
        else:
            return xavier_uniform_(weights, gain=gain)

    def forward(
        self,
        query: torch.Tensor,
        key: torch.Tensor,
        value: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        if self.in_proj_weight is not None:
            if id(query) == id(key):
                # Self attention, get all the projected values at once
                # we compute everything transposed, so that q,k,v stay contiguous after splitting
                qkv = query @ self.in_proj_weight.transpose(-2, -1)

                if self.in_proj_bias is not None:
                    qkv += self.in_proj_bias

                q, k, v = map(
                    lambda x: x.contiguous(),
                    qkv.split(self.out_features, dim=-1),
                )
                return q, k, v

            else:
                # Not self attention
                # - bias free projection
                projections = self.in_proj_weight.split(self.out_features, dim=0)
                q, k, v = map(
                    lambda x, y: x @ y.transpose(1, 0), [query, key, value], projections
                )

                # - optionally add bias
                if self.in_proj_bias is not None:
                    biases = self.in_proj_bias.split(self.out_features, dim=0)
                    q, k, v = map(lambda x, y: x + y, [q, k, v], biases)

                return q, k, v

        # We have a weight per input, but share a bigger bias buffer
        assert (
            self.q_proj_weight is not None
            and self.k_proj_weight is not None
            and self.v_proj_weight is not None
        )

        # - bias free projection
        q, k, v = map(
            lambda x, y: x @ y.transpose(1, 0),
            [query, key, value],
            [self.q_proj_weight, self.k_proj_weight, self.v_proj_weight],
        )

        # - optionally add bias
        if self.in_proj_bias is not None:
            biases = self.in_proj_bias.split(self.out_features, dim=0)
            q, k, v = map(lambda x, y: x + y, [q, k, v], biases)

        return q, k, v
